@@ -1,7 +1,7 @@
 {-|
 Copyright  :  (C) 2013-2016, University of Twente,
                   2016-2017, Myrtle Software Ltd,
-                  2021-2022  QBayLogic B.V.,
+                  2021-2023  QBayLogic B.V.,
                   2022,      Google Inc.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
@@ -50,9 +50,9 @@ import Clash.Annotations.Primitive    (hasBlackBox)
 import Clash.Class.BitPack.Internal.TH (deriveBitPackTuples)
 import Clash.Class.Resize             (zeroExtend, resize)
 import Clash.Promoted.Nat             (SNat(..), snatToNum)
-import Clash.Sized.BitVector          (Bit, BitVector, (++#))
 import Clash.Sized.Internal.BitVector
-  (pack#, split#, checkUnpackUndef, undefined#, unpack#, unsafeToNatural, isLike#)
+  (pack#, split#, checkUnpackUndef, undefined#, unpack#, unsafeToNatural, isLike#,
+   BitVector, Bit, (++#))
 import Clash.XException
 
 {- $setup
@@ -79,6 +79,19 @@ import Clash.XException
 -- instances, as even if @a@ has a statically known size, the length of the
 -- list cannot be known in advance.
 --
+-- It is not possible to give data a custom bit representation by providing a
+-- @BitPack@ instance. A @BitPack@ instance allows no creativity and should
+-- always accurately reflect the bit representation of the data in HDL. You
+-- should always @derive ('Generic', BitPack)@ unless you use a custom data
+-- representation, in which case you should use
+-- 'Clash.Annotations.BitRepresentation.Deriving.deriveBitPack'. Custom
+-- encodings can be created with "Clash.Annotations.BitRepresentation" and
+-- "Clash.Annotations.BitRepresentation.Deriving".
+--
+-- If the @BitPack@ instance does not accurately match the bit representation of
+-- the data in HDL, Clash designs will exhibit incorrect behavior in various
+-- places.
+--
 -- Clash provides some generic functions on packable types in the prelude, such
 -- as indexing into packable stuctures (see "Clash.Class.BitPack.BitIndex") and
 -- bitwise reduction of packable data (see "Clash.Class.BitPack.BitReduction").
@@ -155,7 +168,8 @@ xToBV :: KnownNat n => BitVector n -> BitVector n
 xToBV x =
   unsafeDupablePerformIO (catch (evaluate x)
                                 (\(XException _) -> return undefined#))
-{-# NOINLINE xToBV #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE xToBV #-}
 {-# ANN xToBV hasBlackBox #-}
 
 -- | Pack both arguments to a 'BitVector' and use
@@ -186,7 +200,7 @@ xToBV x =
 -- >>> isLike x1 x3
 -- True
 --
--- __N.B.__: Not synthesizable
+-- __NB__: Not synthesizable
 --
 isLike
   :: (BitPack a)
@@ -298,12 +312,14 @@ instance BitPack Float where
 
 packFloat# :: Float -> BitVector 32
 packFloat# = fromIntegral . floatToWord
-{-# NOINLINE packFloat# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE packFloat# #-}
 {-# ANN packFloat# hasBlackBox #-}
 
 unpackFloat# :: BitVector 32 -> Float
 unpackFloat# (unsafeToNatural -> w) = wordToFloat (fromIntegral w)
-{-# NOINLINE unpackFloat# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unpackFloat# #-}
 {-# ANN unpackFloat# hasBlackBox #-}
 
 instance BitPack Double where
@@ -313,12 +329,14 @@ instance BitPack Double where
 
 packDouble# :: Double -> BitVector 64
 packDouble# = fromIntegral . doubleToWord
-{-# NOINLINE packDouble# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE packDouble# #-}
 {-# ANN packDouble# hasBlackBox #-}
 
 unpackDouble# :: BitVector 64 -> Double
 unpackDouble# (unsafeToNatural -> w) = wordToDouble (fromIntegral w)
-{-# NOINLINE unpackDouble# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unpackDouble# #-}
 {-# ANN unpackDouble# hasBlackBox #-}
 
 instance BitPack CUShort where
@@ -336,7 +354,7 @@ instance BitPack () where
   pack   _ = minBound
   unpack _ = ()
 
--- | __N.B.__: The documentation only shows instances up to /3/-tuples. By
+-- | __NB__: The documentation only shows instances up to /3/-tuples. By
 -- default, instances up to and including /12/-tuples will exist. If the flag
 -- @large-tuples@ is set instances up to the GHC imposed limit will exist. The
 -- GHC imposed limit is either 62 or 64 depending on the GHC version.
@@ -445,6 +463,8 @@ instance GBitPack U1 where
   gUnpack _c _cc _b = U1
 
 -- Instances derived using Generic
+instance BitPack Ordering
+
 instance ( BitPack a
          , BitPack b
          ) => BitPack (Either a b)
