@@ -225,8 +225,10 @@ powUNat x (USucc y) = mulUNat x (powUNat x y)
 -- __NB__: Not synthesizable
 predUNat :: UNat (n+1) -> UNat n
 predUNat (USucc x) = x
+#if __GLASGOW_HASKELL__ != 902
 predUNat UZero     =
   error "predUNat: impossible: 0 minus 1, -1 is not a natural number"
+#endif
 
 -- | Subtract two unary-encoded natural numbers
 --
@@ -267,7 +269,8 @@ infixl 7 `mulSNat`
 -- | Power of two singleton natural numbers
 powSNat :: SNat a -> SNat b -> SNat (a^b)
 powSNat SNat SNat = SNat
-{-# NOINLINE powSNat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE powSNat #-}
 {-# ANN powSNat hasBlackBox #-}
 infixr 8 `powSNat`
 
@@ -295,7 +298,8 @@ flogBaseSNat :: (2 <= base, 1 <= x)
              -> SNat x
              -> SNat (FLog base x)
 flogBaseSNat SNat SNat = SNat
-{-# NOINLINE flogBaseSNat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE flogBaseSNat #-}
 {-# ANN flogBaseSNat hasBlackBox #-}
 
 -- | Ceiling of the logarithm of a natural number
@@ -304,7 +308,8 @@ clogBaseSNat :: (2 <= base, 1 <= x)
              -> SNat x
              -> SNat (CLog base x)
 clogBaseSNat SNat SNat = SNat
-{-# NOINLINE clogBaseSNat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE clogBaseSNat #-}
 {-# ANN clogBaseSNat hasBlackBox #-}
 
 -- | Exact integer logarithm of a natural number
@@ -315,7 +320,8 @@ logBaseSNat :: (FLog base x ~ CLog base x)
             -> SNat x
             -> SNat (Log base x)
 logBaseSNat SNat SNat = SNat
-{-# NOINLINE logBaseSNat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE logBaseSNat #-}
 {-# ANN logBaseSNat hasBlackBox #-}
 
 -- | Power of two of a singleton natural number
@@ -482,7 +488,9 @@ div2Sub1BNat _      = error "div2Sub1BNat: impossible: 2*n+1 ~ 2*n"
 --
 -- __NB__: Not synthesizable
 log2BNat :: BNat (2^n) -> BNat n
+#if __GLASGOW_HASKELL__ != 902
 log2BNat BT = error "log2BNat: log2(0) not defined"
+#endif
 log2BNat (B1 x) = case stripZeros x of
   BT -> BT
   _  -> error "log2BNat: impossible: 2^n ~ 2x+1"
