@@ -2,7 +2,7 @@
 Copyright  :  (C) 2013-2016, University of Twente,
                   2016-2019, Myrtle Software Ltd,
                   2017     , Google Inc.,
-                  2021     , QBayLogic B.V.
+                  2021-2023, QBayLogic B.V.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -106,6 +106,13 @@ module Clash.Signal
   , DomainResetKind
   , DomainInitBehavior
   , DomainResetPolarity
+    -- *** Convenience types
+    -- **** Simplifying
+    -- $conveniencetypes
+
+  , HasSynchronousReset
+  , HasAsynchronousReset
+  , HasDefinedInitialValues
     -- ** Default domains
   , System
   , XilinxSystem
@@ -125,6 +132,7 @@ module Clash.Signal
   , resetPolarity
     -- * Clock
   , Clock
+  , DiffClock
   , periodToHz
   , hzToPeriod
 #ifdef CLASH_MULTIPLE_HIDDEN
@@ -135,10 +143,10 @@ module Clash.Signal
   , Reset
   , unsafeToReset
   , unsafeFromReset
-  , unsafeToHighPolarity
-  , unsafeToLowPolarity
-  , unsafeFromHighPolarity
-  , unsafeFromLowPolarity
+  , unsafeToActiveHigh
+  , unsafeToActiveLow
+  , unsafeFromActiveHigh
+  , unsafeFromActiveLow
 #ifdef CLASH_MULTIPLE_HIDDEN
   , convertReset
 #endif
@@ -261,11 +269,18 @@ module Clash.Signal
   , HiddenClockName
   , HiddenResetName
   , HiddenEnableName
+
+    -- * Deprecated
+  , unsafeFromHighPolarity
+  , unsafeFromLowPolarity
+  , unsafeToHighPolarity
+  , unsafeToLowPolarity
   )
 where
 
 import           Control.Arrow.Transformer.Automaton (Automaton)
 import           GHC.TypeLits          (type (<=))
+import           Data.List             (uncons)
 import           Data.Proxy            (Proxy(..))
 import           Prelude
 import           Test.QuickCheck       (Property, property)
@@ -317,6 +332,34 @@ let countSometimes = s where
 
 -}
 
+{- $conveniencetypes
+
+If you want to write part of your Clash design as domain-polymorphic functions,
+it can be practical to define a design-wide constraint synonym that captures the
+characteristics of the clock domains of the design. Such a constraint synonym
+can be used as a constraint on all domain-polymorphic functions in the design,
+regardless of whether they actually need the constraints from this section.
+
+@
+type DesignDomain dom =
+  ( 'HasSynchronousReset' dom
+  , 'HasDefinedInitialValues' dom
+  )
+
+type DesignDomainHidden dom =
+  ( DesignDomain dom
+  , 'HiddenClockResetEnable' dom
+  )
+
+myFunc ::
+  DesignDomainHidden dom =>
+  'Signal' dom [...]
+@
+
+This way, you don't have to think about which constraints the function you're
+writing has exactly, and the constraint is succinct.
+-}
+
 {- $hiddenclockandreset #hiddenclockandreset#
 Clocks, resets and enables are by default implicitly routed to their components.
 You can see from the type of a component whether it has hidden clock, reset or
@@ -406,7 +449,7 @@ Similarly, there are 'exposeClock', 'exposeReset' and 'exposeEnable' to just
 expose the hidden clock, the hidden reset or the hidden enable argument.
 
 You will need to explicitly apply clocks and resets when you want to use
-components such as PLLs and 'resetSynchronizer':
+components such as PLLs:
 
 @
 topEntity
@@ -415,9 +458,8 @@ topEntity
   -> Signal System Bit
   -> Signal System (BitVector 8)
 topEntity clk rst key1 =
-    let  (pllOut,pllStable) = 'Clash.Intel.ClockGen.altpll' (SSymbol \@\"altpll50\") clk rst
-         rstSync            = 'resetSynchronizer' pllOut (unsafeToHighPolarity pllStable)
-    in   'exposeClockResetEnable' leds pllOut rstSync enableGen
+    let  (pllOut,pllRst) = 'Clash.Intel.ClockGen.altpllSync' clk rst
+    in   'exposeClockResetEnable' leds pllOut pllRst enableGen
   where
     key1R  = isRising 1 key1
     leds   = mealy blinkerT (1, False, 0) key1R
@@ -432,9 +474,8 @@ topEntity
   -> Signal System Bit
   -> Signal System (BitVector 8)
 topEntity clk rst key1 =
-    let  (pllOut,pllStable) = 'Clash.Intel.ClockGen.altpll' (SSymbol \@\"altpll50\") clk rst
-         rstSync            = 'resetSynchronizer' pllOut (unsafeToHighPolarity pllStable)
-    in   'withClockResetEnable' pllOut rstSync enableGen leds
+    let  (pllOut,pllRst) = 'Clash.Intel.ClockGen.altpllSync' clk rst
+    in   'withClockResetEnable' pllOut pllRst enableGen leds
   where
     key1R  = isRising 1 key1
     leds   = mealy blinkerT (1, False, 0) key1R
@@ -1800,7 +1841,8 @@ sample
   -> [a]
 sample s =
   E.sample (exposeClockResetEnable @dom s clockGen resetGen enableGen)
-{-# NOINLINE sample #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE sample #-}
 
 -- | Get a list of /n/ samples from a 'Signal'
 --
@@ -1829,7 +1871,8 @@ sampleN
 sampleN n s0 =
   let s1 = exposeClockResetEnable @dom s0 clockGen resetGen enableGen in
   E.sampleN n s1
-{-# NOINLINE sampleN #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE sampleN #-}
 
 -- | Get an infinite list of samples from a 'Signal', while asserting the reset
 -- line for /m/ clock cycles. 'sampleWithReset' does not return the first /m/
@@ -1850,7 +1893,8 @@ sampleWithReset
 sampleWithReset nReset f0 =
   let f1 = exposeClockResetEnable f0 clockGen (resetGenN @dom nReset) enableGen in
   drop (snatToNum nReset) (E.sample f1)
-{-# NOINLINE sampleWithReset #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE sampleWithReset #-}
 
 -- | Get a list of /n/ samples from a 'Signal', while asserting the reset line
 -- for /m/ clock cycles. 'sampleWithReset' does not return the first /m/ cycles,
@@ -1895,7 +1939,8 @@ sample_lazy
   -> [a]
 sample_lazy s =
   E.sample_lazy (exposeClockResetEnable @dom s clockGen resetGen enableGen)
-{-# NOINLINE sample_lazy #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE sample_lazy #-}
 
 -- | Lazily get a list of /n/ samples from a 'Signal'
 --
@@ -1920,7 +1965,8 @@ sampleN_lazy
   -> [a]
 sampleN_lazy n s =
   E.sampleN_lazy n (exposeClockResetEnable @dom s clockGen resetGen enableGen)
-{-# NOINLINE sampleN_lazy #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE sampleN_lazy #-}
 
 -- * Simulation functions
 
@@ -1951,7 +1997,9 @@ simulate
   -- and/or enable.
   -> [a]
   -> [b]
-simulate f as = simulateWithReset (SNat @1) (head as) f as
+simulate f as = simulateWithReset (SNat @1) rval f as
+  where
+    rval = maybe (error "simulate: no stimuli") fst (uncons as)
 {-# INLINE simulate #-}
 
 -- | Same as 'simulate', but only sample the first /Int/ output values.
@@ -1969,7 +2017,9 @@ simulateN
   -- (and reset)
   -> [a]
   -> [b]
-simulateN n f as = simulateWithResetN (SNat @1) (head as) n f as
+simulateN n f as = simulateWithResetN (SNat @1) rval n f as
+  where
+    rval = maybe (error "simulate: no stimuli") fst (uncons as)
 {-# INLINE simulateN #-}
 
 -- | Same as 'simulate', but with the reset line asserted for /n/ cycles. Similar
@@ -2037,8 +2087,9 @@ simulate_lazy
   -> [b]
 simulate_lazy f0 =
   let f1 = exposeClockResetEnable @dom f0 clockGen resetGen enableGen in
-  tail . E.simulate_lazy f1 . dup1
-{-# NOINLINE simulate_lazy #-}
+  drop 1 . E.simulate_lazy f1 . dup1
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE simulate_lazy #-}
 
 -- | Simulate a (@'Unbundled' a -> 'Unbundled' b@) function given a list of
 -- samples of type @a@
@@ -2063,7 +2114,7 @@ simulateB
   -> [a]
   -> [b]
 simulateB f0 =
-  tail . E.simulateB f1 . dup1
+  drop 1 . E.simulateB f1 . dup1
  where
   f1 =
     withSpecificClockResetEnable
@@ -2073,7 +2124,8 @@ simulateB f0 =
       enableGen
       (const f0)
       (Proxy @dom)
-{-# NOINLINE simulateB #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE simulateB #-}
 
 -- | /Lazily/ simulate a (@'Unbundled' a -> 'Unbundled' b@) function given a
 -- list of samples of type @a@
@@ -2095,7 +2147,7 @@ simulateB_lazy
   -> [a]
   -> [b]
 simulateB_lazy f0 =
-  tail . E.simulateB_lazy f1 . dup1
+  drop 1 . E.simulateB_lazy f1 . dup1
  where
   f1 =
     withSpecificClockResetEnable
@@ -2105,7 +2157,8 @@ simulateB_lazy f0 =
       enableGen
       (const f0)
       (Proxy @dom)
-{-# NOINLINE simulateB_lazy #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE simulateB_lazy #-}
 
 dup1 :: [a] -> [a]
 dup1 (x:xs) = x:x:xs
@@ -2225,7 +2278,7 @@ unsafeSynchronizer =
 --
 -- Example:
 --
--- >>> sampleN @System 8 (unsafeToHighPolarity (holdReset (SNat @2)))
+-- >>> sampleN @System 8 (unsafeToActiveHigh (holdReset (SNat @2)))
 -- [True,True,True,False,False,False,False,False]
 --
 -- 'holdReset' holds the reset for an additional 2 clock cycles for a total
@@ -2243,7 +2296,7 @@ holdReset m =
 
 -- | Like 'fromList', but resets on reset and has a defined reset value.
 --
--- >>> let rst = unsafeFromHighPolarity (fromList [True, True, False, False, True, False])
+-- >>> let rst = unsafeFromActiveHigh (fromList [True, True, False, False, True, False])
 -- >>> let res = withReset rst (fromListWithReset Nothing [Just 'a', Just 'b', Just 'c'])
 -- >>> sampleN @System 6 res
 -- [Nothing,Nothing,Just 'a',Just 'b',Nothing,Just 'a']
@@ -2285,4 +2338,5 @@ signalAutomaton
 signalAutomaton f0 =
   let f1 = exposeClockResetEnable @dom f0 clockGen resetGen enableGen in
   E.signalAutomaton f1
-{-# NOINLINE signalAutomaton #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE signalAutomaton #-}
