@@ -47,6 +47,7 @@ main = do
 -}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE TypeFamilies #-}
 
@@ -96,7 +97,7 @@ import           Clash.Sized.Vector    (Vec, iterateI)
 import qualified Clash.Sized.Vector    as Vector
 import           Clash.Class.BitPack   (BitPack, BitSize, pack, unpack)
 import           Clash.Promoted.Nat    (snatToNum, SNat(..))
-import           Clash.Signal.Internal (sample)
+import           Clash.Signal.Internal (Signal ((:-)), sample)
 import           Clash.XException      (deepseqX, NFDataX)
 import           Clash.Sized.Internal.BitVector
   (BitVector(BV))
@@ -110,7 +111,7 @@ import qualified Data.ByteString.Lazy  as ByteStringLazy
 import           Data.Char             (ord, chr)
 import           Data.IORef
   (IORef, atomicModifyIORef', atomicWriteIORef, newIORef, readIORef)
-import           Data.List             (foldl1', foldl', unzip4, transpose)
+import           Data.List             (foldl1', foldl', unzip4, transpose, uncons)
 import qualified Data.Map.Strict       as Map
 import           Data.Maybe            (fromMaybe, catMaybes)
 import qualified Data.Text             as Text
@@ -140,7 +141,8 @@ type TraceMap  = Map.Map String (TypeRepBS, Period, Width, [Value])
 -- | Map of traces used by the non-internal trace and dumpvcd functions.
 traceMap# :: IORef TraceMap
 traceMap# = unsafePerformIO (newIORef Map.empty)
-{-# NOINLINE traceMap# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceMap# #-}
 
 mkTrace
   :: HasCallStack
@@ -183,7 +185,8 @@ traceSignal# traceMap period traceName signal =
       , signal)
  where
   width = snatToNum (SNat @(BitSize a))
-{-# NOINLINE traceSignal# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceSignal# #-}
 
 -- | Trace a single vector signal: each element in the vector will show up as
 -- a different trace. If the trace name already exists, this function will emit
@@ -209,12 +212,13 @@ traceVecSignal# traceMap period vecTraceName (unbundle -> vecSignal) =
  where
   trace' i s = traceSignal# traceMap period (name' i) s
   name' i    = vecTraceName ++ "_" ++ show i
-{-# NOINLINE traceVecSignal# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceVecSignal# #-}
 
 -- | Trace a single signal. Will emit an error if a signal with the same name
 -- was previously registered.
 --
--- __NB__ Works correctly when creating VCD files from traced signal in
+-- __NB__: Works correctly when creating VCD files from traced signal in
 -- multi-clock circuits. However 'traceSignal1' might be more convenient to
 -- use when the domain of your circuit is polymorphic.
 traceSignal
@@ -230,16 +234,17 @@ traceSignal
   -> Signal dom a
 traceSignal traceName signal =
   case knownDomain @dom of
-    SDomainConfiguration _dom period _edge _reset _init _polarity ->
+    SDomainConfiguration{sPeriod} ->
       unsafePerformIO $
-        traceSignal# traceMap# (snatToNum period) traceName signal
-{-# NOINLINE traceSignal #-}
+        traceSignal# traceMap# (snatToNum sPeriod) traceName signal
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceSignal #-}
 {-# ANN traceSignal hasBlackBox #-}
 
 -- | Trace a single signal. Will emit an error if a signal with the same name
 -- was previously registered.
 --
--- __NB__ associates the traced signal with a clock period of /1/, which
+-- __NB__: Associates the traced signal with a clock period of /1/, which
 -- results in incorrect VCD files when working with circuits that have
 -- multiple clocks. Use 'traceSignal' when working with circuits that have
 -- multiple clocks.
@@ -254,14 +259,15 @@ traceSignal1
   -> Signal dom a
 traceSignal1 traceName signal =
   unsafePerformIO (traceSignal# traceMap# 1 traceName signal)
-{-# NOINLINE traceSignal1 #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceSignal1 #-}
 {-# ANN traceSignal1 hasBlackBox #-}
 
 -- | Trace a single vector signal: each element in the vector will show up as
 -- a different trace. If the trace name already exists, this function will emit
 -- an error.
 --
--- __NB__ Works correctly when creating VCD files from traced signal in
+-- __NB__: Works correctly when creating VCD files from traced signal in
 -- multi-clock circuits. However 'traceSignal1' might be more convenient to
 -- use when the domain of your circuit is polymorphic.
 traceVecSignal
@@ -278,17 +284,18 @@ traceVecSignal
   -> Signal dom (Vec (n+1) a)
 traceVecSignal traceName signal =
   case knownDomain @dom of
-    SDomainConfiguration _dom period _edge _reset _init _polarity ->
+    SDomainConfiguration{sPeriod} ->
       unsafePerformIO $
-        traceVecSignal# traceMap# (snatToNum period) traceName signal
-{-# NOINLINE traceVecSignal #-}
+        traceVecSignal# traceMap# (snatToNum sPeriod) traceName signal
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceVecSignal #-}
 {-# ANN traceVecSignal hasBlackBox #-}
 
 -- | Trace a single vector signal: each element in the vector will show up as
 -- a different trace. If the trace name already exists, this function will emit
 -- an error.
 --
--- __NB__ associates the traced signal with a clock period of /1/, which
+-- __NB__: Associates the traced signal with a clock period of /1/, which
 -- results in incorrect VCD files when working with circuits that have
 -- multiple clocks. Use 'traceSignal' when working with circuits that have
 -- multiple clocks.
@@ -304,7 +311,8 @@ traceVecSignal1
   -> Signal dom (Vec (n+1) a)
 traceVecSignal1 traceName signal =
   unsafePerformIO $ traceVecSignal# traceMap# 1 traceName signal
-{-# NOINLINE traceVecSignal1 #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traceVecSignal1 #-}
 {-# ANN traceVecSignal1 hasBlackBox #-}
 
 iso8601Format :: UTCTime -> String
@@ -340,8 +348,8 @@ dumpVCD## (offset, cycles) traceMap now
       error $ "dumpVCD: no traces found. Extend the given trace names."
   | Map.size traceMap > 126 - 33 =
       Left $ "Tracemap contains more than 93 traces, which is not supported by VCD."
-  | not $ null $ offensiveNames =
-      Left $ unwords [ "Trace '" ++ head offensiveNames ++ "' contains"
+  | (nm:_) <- offensiveNames =
+      Left $ unwords [ "Trace '" ++ nm ++ "' contains"
                      , "non-printable ASCII characters, which is not"
                      , "supported by VCD." ]
   | otherwise =
@@ -406,7 +414,7 @@ dumpVCD## (offset, cycles) traceMap now
   initValues       = map Text.pack $ zipWith ($) formatters inits
 
   formatters = zipWith format widths labels
-  inits = map head valuess'
+  inits = map (maybe (error "dumpVCD##: empty value") fst . uncons) valuess'
   tails = map changed valuess'
 
   -- | Format single value according to VCD spec
@@ -558,14 +566,14 @@ waitForTraces#
   -> IO ()
 waitForTraces# traceMap signal traceNames = do
   atomicWriteIORef traceMap Map.empty
-  rest <- foldM go (sample signal) traceNames
-  return $ deepseqX (head rest) ()
+  rest <- foldM go signal traceNames
+  seq rest (return ())
  where
-  go s nm = do
+  go (s0 :- ss) nm = do
     m <- readIORef traceMap
     if Map.member nm m then
-      return s
+      deepseqX s0 (return ss)
     else
       deepseqX
-        (head s)
-        (go (tail s) nm)
+        s0
+        (go ss nm)
