@@ -1,7 +1,7 @@
 {-|
 Copyright  :  (C) 2013-2016, University of Twente,
                   2016-2019, Myrtle Software Ltd,
-                  2021-2022, QBayLogic B.V.
+                  2021-2023, QBayLogic B.V.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 -}
@@ -112,7 +112,7 @@ import {-# SOURCE #-} Clash.Sized.Internal.BitVector (BitVector (BV), high, low,
 import qualified Clash.Sized.Internal.BitVector as BV
 import Clash.Promoted.Nat         (SNat(..), snatToNum, natToInteger, leToPlusKN)
 import Clash.XException
-  (ShowX (..), NFDataX (..), errorX, showsPrecXWith, rwhnfX)
+  (ShowX (..), NFDataX (..), errorX, showsPrecXWith, rwhnfX, seqX)
 
 {- $setup
 >>> import Clash.Sized.Internal.Index
@@ -120,7 +120,7 @@ import Clash.XException
 
 type role Index nominal
 
--- | Arbitrary-bounded unsigned integer represented by @ceil(log_2(n))@ bits.
+-- | Arbitrarily-bounded unsigned integer represented by @ceil(log_2(n))@ bits
 --
 -- Given an upper bound @n@, an 'Index' @n@ number has a range of: [0 .. @n@-1]
 --
@@ -144,15 +144,20 @@ type role Index nominal
 -- *** Exception: X: Clash.Sized.Index: result 8 is out of bounds: [0..7]
 -- ...
 --
+-- __NB__: The usual Haskell method of converting an integral numeric type to
+-- another, 'fromIntegral', is not well suited for Clash as it will go through
+-- 'Integer' which is arbitrarily bounded in HDL. Instead use
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
+--
 -- Index has the <https://downloads.haskell.org/ghc/latest/docs/html/users_guide/exts/roles.html type role>
 --
 -- >>> :i Index
 -- type role Index nominal
 -- ...
 --
--- as it is not safe to coerce between different range Index. To change the
--- size, use the functions in the 'Clash.Class.Resize.Resize' class.
-#if MIN_VERSION_base(4,15,0)
+-- as it is not safe to coerce between 'Index'es with different ranges. To
+-- change the size, use the functions in the 'Resize' class.
+#if MIN_VERSION_base(4,15,0) && !MIN_VERSION_base(4,17,0)
 data Index (n :: Nat) =
     -- | The constructor, 'I', and the field, 'unsafeToInteger', are not
     -- synthesizable.
@@ -167,7 +172,8 @@ newtype Index (n :: Nat) =
 
 {-# ANN I hasBlackBox #-}
 
-{-# NOINLINE size# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE size# #-}
 size# :: (KnownNat n, 1 <= n) => Index n -> Int
 size# = BV.size# . pack#
 
@@ -186,12 +192,14 @@ instance (KnownNat n, 1 <= n) => BitPack (Index n) where
 fromSNat :: (KnownNat m, n + 1 <= m) => SNat n -> Index m
 fromSNat = snatToNum
 
-{-# NOINLINE pack# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE pack# #-}
 {-# ANN pack# hasBlackBox #-}
 pack# :: Index n -> BitVector (CLog 2 n)
 pack# (I i) = BV 0 (naturalFromInteger i)
 
-{-# NOINLINE unpack# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unpack# #-}
 {-# ANN unpack# hasBlackBox #-}
 unpack# :: (KnownNat n, 1 <= n) => BitVector (CLog 2 n) -> Index n
 unpack# (BV 0 i) = fromInteger_INLINE (naturalToInteger i)
@@ -201,12 +209,14 @@ instance Eq (Index n) where
   (==) = eq#
   (/=) = neq#
 
-{-# NOINLINE eq# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE eq# #-}
 {-# ANN eq# hasBlackBox #-}
 eq# :: (Index n) -> (Index n) -> Bool
 (I n) `eq#` (I m) = n == m
 
-{-# NOINLINE neq# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE neq# #-}
 {-# ANN neq# hasBlackBox #-}
 neq# :: (Index n) -> (Index n) -> Bool
 (I n) `neq#` (I m) = n /= m
@@ -218,16 +228,20 @@ instance Ord (Index n) where
   (<=) = le#
 
 lt#,ge#,gt#,le# :: Index n -> Index n -> Bool
-{-# NOINLINE lt# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE lt# #-}
 {-# ANN lt# hasBlackBox #-}
 lt# (I n) (I m) = n < m
-{-# NOINLINE ge# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE ge# #-}
 {-# ANN ge# hasBlackBox #-}
 ge# (I n) (I m) = n >= m
-{-# NOINLINE gt# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE gt# #-}
 {-# ANN gt# hasBlackBox #-}
 gt# (I n) (I m) = n > m
-{-# NOINLINE le# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE le# #-}
 {-# ANN le# hasBlackBox #-}
 le# (I n) (I m) = n <= m
 
@@ -245,29 +259,35 @@ instance KnownNat n => Enum (Index n) where
 
 toEnum# :: forall n. KnownNat n => Int -> Index n
 toEnum# = fromInteger# . toInteger
-{-# NOINLINE toEnum# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE toEnum# #-}
 {-# ANN toEnum# hasBlackBox #-}
 
 fromEnum# :: forall n. KnownNat n => Index n -> Int
 fromEnum# = fromEnum . toInteger#
-{-# NOINLINE fromEnum# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fromEnum# #-}
 {-# ANN fromEnum# hasBlackBox #-}
 
 enumFrom# :: forall n. KnownNat n => Index n -> [Index n]
 enumFrom# x = [x .. maxBound]
-{-# NOINLINE enumFrom# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFrom# #-}
 
 enumFromThen# :: forall n. KnownNat n => Index n -> Index n -> [Index n]
 enumFromThen# x y = if x <= y then [x, y .. maxBound] else [x, y .. minBound]
-{-# NOINLINE enumFromThen# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromThen# #-}
 
 enumFromTo# :: Index n -> Index n -> [Index n]
 enumFromTo# x y = map I [unsafeToInteger x .. unsafeToInteger y]
-{-# NOINLINE enumFromTo# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromTo# #-}
 
 enumFromThenTo# :: Index n -> Index n -> Index n -> [Index n]
 enumFromThenTo# x1 x2 y = map I [unsafeToInteger x1, unsafeToInteger x2 .. unsafeToInteger y]
-{-# NOINLINE enumFromThenTo# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromThenTo# #-}
 
 instance KnownNat n => Bounded (Index n) where
   minBound = fromInteger# 0
@@ -278,10 +298,15 @@ maxBound# =
   case natToInteger @n of
     0 -> errorX "maxBound of 'Index 0' is undefined"
     n -> fromInteger_INLINE (n - 1)
-{-# NOINLINE maxBound# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE maxBound# #-}
 {-# ANN maxBound# hasBlackBox #-}
 
 -- | Operators report an error on overflow and underflow
+--
+-- __NB__: 'fromInteger'/'fromIntegral' can cause unexpected truncation, as
+-- 'Integer' is arbitrarily bounded during synthesis.  Prefer
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
 instance KnownNat n => Num (Index n) where
   (+)         = (+#)
   (-)         = (-#)
@@ -292,15 +317,18 @@ instance KnownNat n => Num (Index n) where
   fromInteger = fromInteger#
 
 (+#),(-#),(*#) :: KnownNat n => Index n -> Index n -> Index n
-{-# NOINLINE (+#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (+#) #-}
 {-# ANN (+#) hasBlackBox #-}
 (+#) (I a) (I b) = fromInteger_INLINE $ a + b
 
-{-# NOINLINE (-#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (-#) #-}
 {-# ANN (-#) hasBlackBox #-}
 (-#) (I a) (I b) = fromInteger_INLINE $ a - b
 
-{-# NOINLINE (*#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (*#) #-}
 {-# ANN (*#) hasBlackBox #-}
 (*#) (I a) (I b) = fromInteger_INLINE $ a * b
 
@@ -309,7 +337,8 @@ negate# 0 = 0
 negate# i = maxBound -# i +# 1
 
 fromInteger# :: KnownNat n => Integer -> Index n
-{-# NOINLINE fromInteger# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fromInteger# #-}
 {-# ANN fromInteger# hasBlackBox #-}
 fromInteger# = fromInteger_INLINE
 {-# INLINE fromInteger_INLINE #-}
@@ -328,11 +357,13 @@ instance ExtendingNum (Index m) (Index n) where
   mul = times#
 
 plus#, minus# :: Index m -> Index n -> Index (m + n - 1)
-{-# NOINLINE plus# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE plus# #-}
 {-# ANN plus# hasBlackBox #-}
 plus# (I a) (I b) = I (a + b)
 
-{-# NOINLINE minus# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE minus# #-}
 {-# ANN minus# hasBlackBox #-}
 minus# (I a) (I b) =
   let z   = a - b
@@ -341,15 +372,16 @@ minus# (I a) (I b) =
       res = if z < 0 then err else I z
   in  res
 
-{-# NOINLINE times# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE times# #-}
 {-# ANN times# hasBlackBox #-}
 times# :: Index m -> Index n -> Index (((m - 1) * (n - 1)) + 1)
 times# (I a) (I b) = I (a * b)
 
 instance (KnownNat n, 1 <= n) => SaturatingNum (Index n) where
-  satAdd SatWrap !a !b =
+  satAdd SatWrap a b =
     case natToInteger @n of
-      1 -> fromInteger# 0
+      1 -> a +# b
       _ -> leToPlusKN @1 @n $
         case plus# a b of
           z | let m = fromInteger# (natToInteger @n)
@@ -387,9 +419,9 @@ instance (KnownNat n, 1 <= n) => SaturatingNum (Index n) where
        then fromInteger# 0
        else a -# b
 
-  satMul SatWrap !a !b =
+  satMul SatWrap a b =
     case natToInteger @n of
-      1 -> fromInteger# 0
+      1 -> a *# b
       2 -> case a of {0 -> 0; _ -> b}
       _ -> leToPlusKN @1 @n $
         case times# a b of
@@ -414,9 +446,9 @@ instance (KnownNat n, 1 <= n) => SaturatingNum (Index n) where
           , z > m -> maxBound#
         z -> resize# z
 
-  satSucc SatError !a =
+  satSucc SatError a =
     case natToInteger @n of
-      1 -> errorX "Index.satSucc: overflow"
+      1 -> a `seqX` errorX "Index.satSucc: overflow"
       _ -> satAdd SatError a $ fromInteger# 1
   satSucc satMode !a =
     case natToInteger @n of
@@ -424,9 +456,9 @@ instance (KnownNat n, 1 <= n) => SaturatingNum (Index n) where
       _ -> satAdd satMode a $ fromInteger# 1
   {-# INLINE satSucc #-}
 
-  satPred SatError !a =
+  satPred SatError a =
     case natToInteger @n of
-      1 -> errorX "Index.satPred: underflow"
+      1 -> a `seqX` errorX "Index.satPred: underflow"
       _ -> satSub SatError a $ fromInteger# 1
   satPred satMode !a =
     case natToInteger @n of
@@ -437,6 +469,9 @@ instance (KnownNat n, 1 <= n) => SaturatingNum (Index n) where
 instance KnownNat n => Real (Index n) where
   toRational = toRational . toInteger#
 
+-- | __NB__: 'toInteger'/'fromIntegral' can cause unexpected truncation, as
+-- 'Integer' is arbitrarily bounded during synthesis.  Prefer
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
 instance KnownNat n => Integral (Index n) where
   quot        = quot#
   rem         = rem#
@@ -447,14 +482,17 @@ instance KnownNat n => Integral (Index n) where
   toInteger   = toInteger#
 
 quot#,rem# :: Index n -> Index n -> Index n
-{-# NOINLINE quot# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE quot# #-}
 {-# ANN quot# hasBlackBox #-}
 (I a) `quot#` (I b) = I (a `div` b)
-{-# NOINLINE rem# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rem# #-}
 {-# ANN rem# hasBlackBox #-}
 (I a) `rem#` (I b) = I (a `rem` b)
 
-{-# NOINLINE toInteger# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE toInteger# #-}
 {-# ANN toInteger# hasBlackBox #-}
 toInteger# :: Index n -> Integer
 toInteger# (I n) = n
@@ -498,7 +536,8 @@ instance Resize Index where
 
 resize# :: KnownNat m => Index n -> Index m
 resize# (I i) = fromInteger_INLINE i
-{-# NOINLINE resize# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE resize# #-}
 {-# ANN resize# hasBlackBox #-}
 
 instance KnownNat n => Lift (Index n) where
