@@ -1,7 +1,7 @@
 {-|
 Copyright  :  (C) 2013-2016, University of Twente,
                   2016     , Myrtle Software Ltd,
-                  2021-2022, QBayLogic B.V.
+                  2021-2023, QBayLogic B.V.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 -}
@@ -92,7 +92,11 @@ import Data.Default.Class             (Default (..))
 import Data.Proxy                     (Proxy (..))
 import Text.Read                      (Read (..), ReadPrec)
 import Text.Printf                    (PrintfArg (..), printf)
+#if MIN_VERSION_base(4,16,0)
+import GHC.Exts                       (wordToWord8#, wordToWord16#, wordToWord32#)
+#else
 import GHC.Exts                       (narrow8Word#, narrow16Word#, narrow32Word#)
+#endif
 import GHC.Generics                   (Generic)
 #if MIN_VERSION_base(4,15,0)
 import GHC.Num.BigNat                 (bigNatToWord, bigNatToWord#)
@@ -156,7 +160,11 @@ type role Unsigned nominal
 --
 -- Given @n@ bits, an 'Unsigned' @n@ number has a range of: [0 .. 2^@n@-1]
 --
--- __NB__: The 'Num' operators perform @wrap-around@ on overflow. If you want
+-- * __NB__: The usual Haskell method of converting an integral numeric type to
+-- another, 'fromIntegral', is not well suited for Clash as it will go through
+-- 'Integer' which is arbitrarily bounded in HDL. Instead use
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
+-- * __NB__: The 'Num' operators perform @wrap-around@ on overflow. If you want
 -- saturation on overflow, check out the 'SaturatingNum' class.
 --
 -- >>> maxBound :: Unsigned 3
@@ -192,7 +200,7 @@ type role Unsigned nominal
 --
 -- as it is not safe to coerce between different width Unsigned. To change the
 -- width, use the functions in the 'Clash.Class.Resize.Resize' class.
-#if MIN_VERSION_base(4,15,0)
+#if MIN_VERSION_base(4,15,0) && !MIN_VERSION_base(4,17,0)
 data Unsigned (n :: Nat) =
     -- | The constructor, 'U', and the field, 'unsafeToNatural', are not
     -- synthesizable.
@@ -207,7 +215,8 @@ newtype Unsigned (n :: Nat) =
 
 {-# ANN U hasBlackBox #-}
 
-{-# NOINLINE size# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE size# #-}
 {-# ANN size# hasBlackBox #-}
 size# :: KnownNat n => Unsigned n -> Int
 #if MIN_VERSION_base(4,15,0)
@@ -242,12 +251,14 @@ instance KnownNat n => BitPack (Unsigned n) where
   pack   = packXWith pack#
   unpack = unpack#
 
-{-# NOINLINE pack# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE pack# #-}
 {-# ANN pack# hasBlackBox #-}
 pack# :: Unsigned n -> BitVector n
 pack# (U i) = BV 0 i
 
-{-# NOINLINE unpack# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unpack# #-}
 {-# ANN unpack# hasBlackBox #-}
 unpack# :: KnownNat n => BitVector n -> Unsigned n
 unpack# (BV 0 i) = U i
@@ -257,12 +268,14 @@ instance Eq (Unsigned n) where
   (==) = eq#
   (/=) = neq#
 
-{-# NOINLINE eq# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE eq# #-}
 {-# ANN eq# hasBlackBox #-}
 eq# :: Unsigned n -> Unsigned n -> Bool
 eq# (U v1) (U v2) = v1 == v2
 
-{-# NOINLINE neq# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE neq# #-}
 {-# ANN neq# hasBlackBox #-}
 neq# :: Unsigned n -> Unsigned n -> Bool
 neq# (U v1) (U v2) = v1 /= v2
@@ -274,16 +287,20 @@ instance Ord (Unsigned n) where
   (<=) = le#
 
 lt#,ge#,gt#,le# :: Unsigned n -> Unsigned n -> Bool
-{-# NOINLINE lt# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE lt# #-}
 {-# ANN lt# hasBlackBox #-}
 lt# (U n) (U m) = n < m
-{-# NOINLINE ge# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE ge# #-}
 {-# ANN ge# hasBlackBox #-}
 ge# (U n) (U m) = n >= m
-{-# NOINLINE gt# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE gt# #-}
 {-# ANN gt# hasBlackBox #-}
 gt# (U n) (U m) = n > m
-{-# NOINLINE le# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE le# #-}
 {-# ANN le# hasBlackBox #-}
 le# (U n) (U m) = n <= m
 
@@ -300,10 +317,9 @@ instance KnownNat n => Enum (Unsigned n) where
 
   pred n
     | n == minBound =
-        error $ "'pred' was called on (" <> show @(Unsigned n) maxBound <> " :: "
-             <> "Unsigned " <> show (natToNatural @n) <> ") and caused an "
-             <> "underflow. Use 'satPred' and specify a SaturationMode if you "
-             <> "need other behavior."
+        error $ "'pred' was called on (0 :: Unsigned " <> show (natToNatural @n)
+             <> ") and caused an overflow. Use 'satPred' and specify a "
+             <> "SaturationMode if you need other behavior."
     | otherwise = n -# fromInteger# 1
 
   toEnum         = toEnum#
@@ -315,12 +331,14 @@ instance KnownNat n => Enum (Unsigned n) where
 
 toEnum# :: forall n. KnownNat n => Int -> Unsigned n
 toEnum# = fromInteger# . toInteger
-{-# NOINLINE toEnum# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE toEnum# #-}
 {-# ANN toEnum# hasBlackBox #-}
 
 fromEnum# :: forall n. KnownNat n => Unsigned n -> Int
 fromEnum# = fromEnum . toInteger#
-{-# NOINLINE fromEnum# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fromEnum# #-}
 {-# ANN fromEnum# hasBlackBox #-}
 
 enumFrom# :: forall n. KnownNat n => Unsigned n -> [Unsigned n]
@@ -330,7 +348,8 @@ enumFrom# = \x -> map (U . (`mod` m)) [unsafeToNatural x .. unsafeToNatural (max
 #else
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
-{-# NOINLINE enumFrom# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFrom# #-}
 
 enumFromThen# :: forall n. KnownNat n => Unsigned n -> Unsigned n -> [Unsigned n]
 enumFromThen# = \x y -> toUnsigneds [unsafeToNatural x, unsafeToNatural y .. bound x y]
@@ -342,7 +361,8 @@ enumFromThen# = \x y -> toUnsigneds [unsafeToNatural x, unsafeToNatural y .. bou
 #else
   m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
-{-# NOINLINE enumFromThen# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromThen# #-}
 
 enumFromTo# :: forall n. KnownNat n => Unsigned n -> Unsigned n -> [Unsigned n]
 enumFromTo# = \x y -> map (U . (`mod` m)) [unsafeToNatural x .. unsafeToNatural y]
@@ -351,7 +371,8 @@ enumFromTo# = \x y -> map (U . (`mod` m)) [unsafeToNatural x .. unsafeToNatural
 #else
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
-{-# NOINLINE enumFromTo# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromTo# #-}
 
 enumFromThenTo# :: forall n. KnownNat n => Unsigned n -> Unsigned n -> Unsigned n -> [Unsigned n]
 enumFromThenTo# = \x1 x2 y -> map (U . (`mod` m)) [unsafeToNatural x1, unsafeToNatural x2 .. unsafeToNatural y]
@@ -360,7 +381,8 @@ enumFromThenTo# = \x1 x2 y -> map (U . (`mod` m)) [unsafeToNatural x1, unsafeToN
 #else
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
-{-# NOINLINE enumFromThenTo# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE enumFromThenTo# #-}
 
 instance KnownNat n => Bounded (Unsigned n) where
   minBound = minBound#
@@ -368,14 +390,19 @@ instance KnownNat n => Bounded (Unsigned n) where
 
 minBound# :: Unsigned n
 minBound# = U 0
-{-# NOINLINE minBound# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE minBound# #-}
 {-# ANN minBound# hasBlackBox #-}
 
 maxBound# :: forall n. KnownNat n => Unsigned n
 maxBound# = let m = 1 `shiftL` (natToNum @n) in  U (m - 1)
-{-# NOINLINE maxBound# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE maxBound# #-}
 {-# ANN maxBound# hasBlackBox #-}
 
+-- | __NB__: 'fromInteger'/'fromIntegral' can cause unexpected truncation, as
+-- 'Integer' is arbitrarily bounded during synthesis.  Prefer
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
 instance KnownNat n => Num (Unsigned n) where
   (+)         = (+#)
   (-)         = (-#)
@@ -386,7 +413,8 @@ instance KnownNat n => Num (Unsigned n) where
   fromInteger = fromInteger#
 
 (+#),(-#),(*#) :: forall n . KnownNat n => Unsigned n -> Unsigned n -> Unsigned n
-{-# NOINLINE (+#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (+#) #-}
 {-# ANN (+#) hasBlackBox #-}
 (+#) = \(U i) (U j) -> U (addMod m i j)
 #if MIN_VERSION_base(4,15,0)
@@ -395,7 +423,8 @@ instance KnownNat n => Num (Unsigned n) where
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
 
-{-# NOINLINE (-#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (-#) #-}
 {-# ANN (-#) hasBlackBox #-}
 (-#) = \(U i) (U j) -> U (subMod m i j)
 #if MIN_VERSION_base(4,15,0)
@@ -404,7 +433,8 @@ instance KnownNat n => Num (Unsigned n) where
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
 
-{-# NOINLINE (*#) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (*#) #-}
 {-# ANN (*#) hasBlackBox #-}
 (*#) = \(U i) (U j) -> U (mulMod2 m i j)
 #if MIN_VERSION_base(4,15,0)
@@ -413,7 +443,8 @@ instance KnownNat n => Num (Unsigned n) where
   where m = (1 `shiftL` fromInteger (natVal (Proxy @n))) - 1
 #endif
 
-{-# NOINLINE negate# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE negate# #-}
 {-# ANN negate# hasBlackBox #-}
 negate# :: forall n . KnownNat n => Unsigned n -> Unsigned n
 negate# = \(U i) -> U (negateMod m i)
@@ -423,7 +454,8 @@ negate# = \(U i) -> U (negateMod m i)
   where m = 1 `shiftL` fromInteger (natVal (Proxy @n))
 #endif
 
-{-# NOINLINE fromInteger# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fromInteger# #-}
 {-# ANN fromInteger# hasBlackBox #-}
 fromInteger# :: forall n . KnownNat n => Integer -> Unsigned n
 #if MIN_VERSION_base(4,15,0)
@@ -443,12 +475,14 @@ instance (KnownNat m, KnownNat n) => ExtendingNum (Unsigned m) (Unsigned n) wher
   type MResult (Unsigned m) (Unsigned n) = Unsigned (m + n)
   mul = times#
 
-{-# NOINLINE plus# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE plus# #-}
 {-# ANN plus# hasBlackBox #-}
 plus# :: Unsigned m -> Unsigned n -> Unsigned (Max m n + 1)
 plus# (U a) (U b) = U (a + b)
 
-{-# NOINLINE minus# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE minus# #-}
 {-# ANN minus# hasBlackBox #-}
 minus# :: forall m n . (KnownNat m, KnownNat n) => Unsigned m -> Unsigned n
                                                 -> Unsigned (Max m n + 1)
@@ -462,7 +496,8 @@ minus# = \(U a) (U b) -> U (subMod mask a b)
   mask = 1 `shiftL` sz
 #endif
 
-{-# NOINLINE times# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE times# #-}
 {-# ANN times# hasBlackBox #-}
 times# :: Unsigned m -> Unsigned n -> Unsigned (m + n)
 times# (U a) (U b) = U (a * b)
@@ -470,6 +505,9 @@ times# (U a) (U b) = U (a * b)
 instance KnownNat n => Real (Unsigned n) where
   toRational = toRational . toInteger#
 
+-- | __NB__: 'toInteger'/'fromIntegral' can cause unexpected truncation, as
+-- 'Integer' is arbitrarily bounded during synthesis.  Prefer
+-- 'Clash.Class.BitPack.bitCoerce' and the 'Resize' class.
 instance KnownNat n => Integral (Unsigned n) where
   quot        = quot#
   rem         = rem#
@@ -480,14 +518,17 @@ instance KnownNat n => Integral (Unsigned n) where
   toInteger   = toInteger#
 
 quot#,rem# :: Unsigned n -> Unsigned n -> Unsigned n
-{-# NOINLINE quot# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE quot# #-}
 {-# ANN quot# hasBlackBox #-}
 quot# (U i) (U j) = U (i `quot` j)
-{-# NOINLINE rem# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rem# #-}
 {-# ANN rem# hasBlackBox #-}
 rem# (U i) (U j) = U (i `rem` j)
 
-{-# NOINLINE toInteger# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE toInteger# #-}
 {-# ANN toInteger# hasBlackBox #-}
 toInteger# :: Unsigned n -> Integer
 toInteger# (U i) = naturalToInteger i
@@ -519,29 +560,34 @@ instance KnownNat n => Bits (Unsigned n) where
   rotateR v i       = rotateR# v i
   popCount u        = popCount (pack# u)
 
-{-# NOINLINE and# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE and# #-}
 {-# ANN and# hasBlackBox #-}
 and# :: Unsigned n -> Unsigned n -> Unsigned n
 and# (U v1) (U v2) = U (v1 .&. v2)
 
-{-# NOINLINE or# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE or# #-}
 {-# ANN or# hasBlackBox #-}
 or# :: Unsigned n -> Unsigned n -> Unsigned n
 or# (U v1) (U v2) = U (v1 .|. v2)
 
-{-# NOINLINE xor# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE xor# #-}
 {-# ANN xor# hasBlackBox #-}
 xor# :: Unsigned n -> Unsigned n -> Unsigned n
 xor# (U v1) (U v2) = U (v1 `xor` v2)
 
-{-# NOINLINE complement# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE complement# #-}
 {-# ANN complement# hasBlackBox #-}
 complement# :: forall n . KnownNat n => Unsigned n -> Unsigned n
 complement# = \(U i) -> U (complementN i)
   where complementN = complementMod (natVal (Proxy @n))
 
 shiftL#, shiftR#, rotateL#, rotateR# :: forall n .KnownNat n => Unsigned n -> Int -> Unsigned n
-{-# NOINLINE shiftL# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE shiftL# #-}
 {-# ANN shiftL# hasBlackBox #-}
 shiftL# = \(U v) i ->
 #if MIN_VERSION_base(4,15,0)
@@ -561,7 +607,8 @@ shiftL# = \(U v) i ->
   m  = 1 `shiftL` sz
 #endif
 
-{-# NOINLINE shiftR# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE shiftR# #-}
 {-# ANN shiftR# hasBlackBox #-}
 -- shiftR# doesn't need the KnownNat constraint
 -- But having the same type signature for all shift and rotate functions
@@ -571,7 +618,8 @@ shiftR# (U v) i
               $ "'shiftR' undefined for negative number: " ++ show i
   | otherwise = U (shiftR v i)
 
-{-# NOINLINE rotateL# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rotateL# #-}
 {-# ANN rotateL# hasBlackBox #-}
 rotateL# =
   \(U n) b ->
@@ -598,7 +646,8 @@ rotateL# =
     m  = 1 `shiftL` sz
 #endif
 
-{-# NOINLINE rotateR# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rotateR# #-}
 {-# ANN rotateR# hasBlackBox #-}
 rotateR# =
   \(U n) b ->
@@ -636,7 +685,8 @@ instance Resize Unsigned where
   zeroExtend = extend
   truncateB  = resize#
 
-{-# NOINLINE resize# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE resize# #-}
 {-# ANN resize# hasBlackBox #-}
 resize# :: forall n m . KnownNat m => Unsigned n -> Unsigned m
 resize# = \(U i) -> if i >= m then U (i `mod` m) else U i
@@ -687,7 +737,7 @@ instance KnownNat n => SaturatingNum (Unsigned n) where
     let r = minus# a b
     in  case msb r of
           0 -> resize# r
-          _ -> errorX "Unsigned.satSub: underflow"
+          _ -> errorX "Unsigned.satSub: overflow"
   satSub _ a b =
     let r = minus# a b
     in  case msb r of
@@ -723,7 +773,7 @@ instance KnownNat n => SaturatingNum (Unsigned n) where
   {-# INLINE satSucc #-}
 
   satPred SatError a
-    | a == minBound = errorX "Unsigned.satPred: underflow"
+    | a == minBound = errorX "Unsigned.satPred: overflow"
   satPred satMode a = satSub satMode a 1
   {-# INLINE satPred #-}
 
@@ -755,40 +805,53 @@ unsignedToWord (U (NB u#)) = bigNatToWord u#
 unsignedToWord (U (NatS# u#)) = W# u#
 unsignedToWord (U (NatJ# u#)) = W# (bigNatToWord u#)
 #endif
-{-# NOINLINE unsignedToWord #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unsignedToWord #-}
 {-# ANN unsignedToWord hasBlackBox #-}
 
 unsigned8toWord8 :: Unsigned 8 -> Word8
-#if MIN_VERSION_base(4,15,0)
+#if MIN_VERSION_base(4,16,0)
+unsigned8toWord8 (U (NS u#)) = W8# (wordToWord8# u#)
+unsigned8toWord8 (U (NB u#)) = W8# (wordToWord8# (bigNatToWord# u#))
+#elif MIN_VERSION_base(4,15,0)
 unsigned8toWord8 (U (NS u#)) = W8# (narrow8Word# u#)
 unsigned8toWord8 (U (NB u#)) = W8# (narrow8Word# (bigNatToWord# u#))
 #else
 unsigned8toWord8 (U (NatS# u#)) = W8# (narrow8Word# u#)
 unsigned8toWord8 (U (NatJ# u#)) = W8# (narrow8Word# (bigNatToWord u#))
 #endif
-{-# NOINLINE unsigned8toWord8 #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unsigned8toWord8 #-}
 {-# ANN unsigned8toWord8 hasBlackBox #-}
 
 unsigned16toWord16 :: Unsigned 16 -> Word16
-#if MIN_VERSION_base(4,15,0)
+#if MIN_VERSION_base(4,16,0)
+unsigned16toWord16 (U (NS u#)) = W16# (wordToWord16# u#)
+unsigned16toWord16 (U (NB u#)) = W16# (wordToWord16# (bigNatToWord# u#))
+#elif MIN_VERSION_base(4,15,0)
 unsigned16toWord16 (U (NS u#)) = W16# (narrow16Word# u#)
 unsigned16toWord16 (U (NB u#)) = W16# (narrow16Word# (bigNatToWord# u#))
 #else
 unsigned16toWord16 (U (NatS# u#)) = W16# (narrow16Word# u#)
 unsigned16toWord16 (U (NatJ# u#)) = W16# (narrow16Word# (bigNatToWord u#))
 #endif
-{-# NOINLINE unsigned16toWord16 #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unsigned16toWord16 #-}
 {-# ANN unsigned16toWord16 hasBlackBox #-}
 
 unsigned32toWord32 :: Unsigned 32 -> Word32
-#if MIN_VERSION_base(4,15,0)
+#if MIN_VERSION_base(4,16,0)
+unsigned32toWord32 (U (NS u#)) = W32# (wordToWord32# u#)
+unsigned32toWord32 (U (NB u#)) = W32# (wordToWord32# (bigNatToWord# u#))
+#elif MIN_VERSION_base(4,15,0)
 unsigned32toWord32 (U (NS u#)) = W32# (narrow32Word# u#)
 unsigned32toWord32 (U (NB u#)) = W32# (narrow32Word# (bigNatToWord# u#))
 #else
 unsigned32toWord32 (U (NatS# u#)) = W32# (narrow32Word# u#)
 unsigned32toWord32 (U (NatJ# u#)) = W32# (narrow32Word# (bigNatToWord u#))
 #endif
-{-# NOINLINE unsigned32toWord32 #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unsigned32toWord32 #-}
 {-# ANN unsigned32toWord32 hasBlackBox #-}
 
 {-# RULES
