@@ -57,7 +57,9 @@ module Clash.Sized.RTree
   )
 where
 
+#if !MIN_VERSION_base(4,18,0)
 import Control.Applicative         (liftA2)
+#endif
 import Control.DeepSeq             (NFData(..))
 import qualified Control.Lens      as Lens
 import Data.Default.Class          (Default (..))
@@ -122,14 +124,20 @@ instance NFData a => NFData (RTree d a) where
 
 textract :: RTree 0 a -> a
 textract (RLeaf x)   = x
+#if __GLASGOW_HASKELL__ != 902
 textract (RBranch _ _) = error $ "textract: nodes hold no values"
-{-# NOINLINE textract #-}
+#endif
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE textract #-}
 {-# ANN textract hasBlackBox #-}
 
 tsplit :: RTree (d+1) a -> (RTree d a,RTree d a)
 tsplit (RBranch l r) = (l,r)
+#if __GLASGOW_HASKELL__ != 902
 tsplit (RLeaf _)   = error $ "tsplit: leaf is atomic"
-{-# NOINLINE tsplit #-}
+#endif
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE tsplit #-}
 {-# ANN tsplit hasBlackBox #-}
 
 -- | RLeaf of a perfect depth tree
@@ -267,7 +275,7 @@ tree-structure of adders:
 @
 populationCount :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
                 => BitVector (2^d) -> Index (2^d+1)
-populationCount = tfold fromIntegral (+) . v2t . bv2v
+populationCount = tfold (resize . bv2i . pack) (+) . v2t . bv2v
 @
 
 The \"problem\" with this description is that all adders have the same
@@ -289,12 +297,12 @@ type:
 We have such an adder in the form of the 'Clash.Class.Num.add' function, as
 defined in the instance 'Clash.Class.Num.ExtendingNum' instance of 'Index'.
 However, we cannot simply use 'Clash.Sized.Vector.fold' to create a tree-structure of
-'Clash.Class.Num.add'es:
+'Clash.Class.Num.add's:
 #if __GLASGOW_HASKELL__ >= 900
 >>> :{
 let populationCount' :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
                      => BitVector (2^d) -> Index (2^d+1)
-    populationCount' = tfold fromIntegral add . v2t . bv2v
+    populationCount' = tfold (resize . bv2i . pack) add . v2t . bv2v
 :}
 <BLANKLINE>
 <interactive>:...
@@ -306,8 +314,9 @@ let populationCount' :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
                 -> Index ((2 ^ d) + 1)
                 -> AResult (Index ((2 ^ d) + 1)) (Index ((2 ^ d) + 1))
     • In the second argument of ‘tfold’, namely ‘add’
-      In the first argument of ‘(.)’, namely ‘tfold fromIntegral add’
-      In the expression: tfold fromIntegral add . v2t . bv2v
+      In the first argument of ‘(.)’, namely
+        ‘tfold (resize . bv2i . pack) add’
+      In the expression: tfold (resize . bv2i . pack) add . v2t . bv2v
     • Relevant bindings include
         populationCount' :: BitVector (2 ^ d) -> Index ((2 ^ d) + 1)
           (bound at ...)
@@ -316,7 +325,7 @@ let populationCount' :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
 >>> :{
 let populationCount' :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
                      => BitVector (2^d) -> Index (2^d+1)
-    populationCount' = tfold fromIntegral add . v2t . bv2v
+    populationCount' = tfold (resize . bv2i . pack) add . v2t . bv2v
 :}
 <BLANKLINE>
 <interactive>:...
@@ -328,8 +337,9 @@ let populationCount' :: (KnownNat (2^d), KnownNat d, KnownNat (2^d+1))
                      -> Index ((2 ^ d) + 1)
                      -> AResult (Index ((2 ^ d) + 1)) (Index ((2 ^ d) + 1))
     • In the second argument of ‘tfold’, namely ‘add’
-      In the first argument of ‘(.)’, namely ‘tfold fromIntegral add’
-      In the expression: tfold fromIntegral add . v2t . bv2v
+      In the first argument of ‘(.)’, namely
+        ‘tfold (resize . bv2i . pack) add’
+      In the expression: tfold (resize . bv2i . pack) add . v2t . bv2v
     • Relevant bindings include
         populationCount' :: BitVector (2 ^ d) -> Index ((2 ^ d) + 1)
           (bound at ...)
@@ -346,7 +356,6 @@ the form of 'dtfold':
 @
 {\-\# LANGUAGE UndecidableInstances \#-\}
 import Data.Singletons
-import Data.Proxy
 
 data IIndex (f :: 'TyFun' Nat Type) :: Type
 type instance 'Apply' IIndex l = 'Index' ((2^l)+1)
@@ -354,7 +363,7 @@ type instance 'Apply' IIndex l = 'Index' ((2^l)+1)
 populationCount' :: (KnownNat k, KnownNat (2^k))
                  => BitVector (2^k) -> Index ((2^k)+1)
 populationCount' bv = 'tdfold' (Proxy @IIndex)
-                             fromIntegral
+                             (resize . bv2i . pack)
                              (\\_ x y -> 'Clash.Class.Num.add' x y)
                              ('v2t' ('Clash.Sized.Vector.bv2v' bv))
 @
@@ -372,7 +381,7 @@ tdfold :: forall p k a . KnownNat k
        -> (forall l . SNat l -> (p @@ l) -> (p @@ l) -> (p @@ (l+1)))
        -- ^ Function to fold the branches with.
        --
-       -- __NB:__ @SNat l@ is the depth of the two sub-branches.
+       -- __NB__: @SNat l@ is the depth of the two sub-branches.
        -> RTree k a -- ^ Tree to fold over.
        -> (p @@ k)
 tdfold _ f g = go SNat
@@ -381,7 +390,8 @@ tdfold _ f g = go SNat
     go _  (RLeaf a)   = f a
     go sn (RBranch l r) = let sn' = sn `subSNat` d1
                       in  g sn' (go sn' l) (go sn' r)
-{-# NOINLINE tdfold #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE tdfold #-}
 {-# ANN tdfold hasBlackBox #-}
 
 data TfoldTree (a :: Type) (f :: TyFun Nat Type) :: Type
@@ -410,7 +420,8 @@ treplicate sn a = go (toUNat sn)
     go :: UNat n -> RTree n a
     go UZero      = LR a
     go (USucc un) = BR (go un) (go un)
-{-# NOINLINE treplicate #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE treplicate #-}
 {-# ANN treplicate hasBlackBox #-}
 
 -- | \"'trepeat' @a@\" creates a tree with as many copies of /a/ as demanded by
