@@ -1,7 +1,7 @@
 {-|
 Copyright  :  (C) 2013-2016, University of Twente,
                   2017     , Myrtle Software Ltd
-                  2022     , QBayLogic B.V.
+                  2022-2023, QBayLogic B.V.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 -}
@@ -75,9 +75,9 @@ module Clash.Sized.Vector
   , foldr, foldl, foldr1, foldl1, fold
   , ifoldr, ifoldl
     -- ** Specialized folds
-  , dfold, dtfold, vfold
+  , dfold, dtfold, vfold, maximum, minimum
     -- * Prefix sums (scans)
-  , scanl, scanr, postscanl, postscanr
+  , scanl, scanl1, scanr, scanr1, postscanl, postscanr
   , mapAccumL, mapAccumR
     -- * Stencil computations
   , stencil1d, stencil2d
@@ -107,6 +107,9 @@ import Data.Constraint.Nat        (leZero)
 import Data.Data
   (Data (..), Constr, DataType, Fixity (..), Typeable, mkConstr, mkDataType)
 import Data.Either                (isLeft)
+#if MIN_VERSION_base(4,18,0)
+import qualified Data.Foldable1   as F1
+#endif
 import Data.Default.Class         (Default (..))
 import qualified Data.Foldable    as F
 import Data.Kind                  (Type)
@@ -127,8 +130,9 @@ import Prelude                    hiding ((++), (!!), concat, concatMap, drop,
                                           foldl, foldl1, foldr, foldr1, head,
                                           init, iterate, last, length, map,
                                           repeat, replicate, reverse, scanl,
-                                          scanr, splitAt, tail, take, unzip,
-                                          unzip3, zip, zip3, zipWith, zipWith3)
+                                          scanl1, scanr, scanr1, splitAt, tail,
+                                          take, unzip, unzip3, zip, zip3, zipWith,
+                                          zipWith3, maximum, minimum)
 import qualified Data.String.Interpolate as I
 import qualified Prelude          as P
 import Test.QuickCheck
@@ -136,10 +140,14 @@ import Test.QuickCheck
 import Unsafe.Coerce              (unsafeCoerce)
 
 import Clash.Annotations.Primitive
-  (Primitive(InlinePrimitive), HDL(..), dontTranslate, hasBlackBox)
+  (Primitive(InlineYamlPrimitive), HDL(..), dontTranslate, hasBlackBox)
+import Clash.Magic (clashCompileError)
 import Clash.Promoted.Nat
-  (SNat (..), SNatLE (..), UNat (..), compareSNat, leToPlus, pow2SNat,
+  (SNat (..), SNatLE (..), UNat (..), compareSNat, pow2SNat,
    snatProxy, snatToInteger, subSNat, withSNat, toUNat, natToInteger)
+#if MIN_VERSION_base(4,18,0)
+import Clash.Promoted.Nat (leToPlus)
+#endif
 import Clash.Promoted.Nat.Literals (d1)
 import Clash.Sized.Internal.BitVector (Bit, BitVector (..), split#)
 import Clash.Sized.Index          (Index)
@@ -152,6 +160,9 @@ import Clash.XException           (ShowX (..), NFDataX (..), seqX, isX)
 >>> :set -XTypeOperators
 >>> :set -XTemplateHaskell
 >>> :set -XFlexibleContexts
+>>> :set -fplugin GHC.TypeLits.Normalise
+>>> :set -fplugin GHC.TypeLits.KnownNat.Solver
+>>> :set -fplugin GHC.TypeLits.Extra.Solver
 >>> :m -Prelude
 >>> import Clash.Prelude
 >>> import qualified Clash.Sized.Vector as Vec
@@ -216,8 +227,8 @@ instance (KnownNat n, Typeable a, Data a) => Data (Vec n a) where
   gfoldl f z xs = case compareSNat (SNat @n) (SNat @0) of
     SNatLE -> case leZero @n of
                   Sub Dict -> z Nil
-    SNatGT -> let (y :> ys) = xs
-              in (z @(a -> Vec (n-1) a -> Vec n a) (:>) `f` y `f` ys)
+    SNatGT -> case xs of
+                  (y :> ys) -> (z @(a -> Vec (n-1) a -> Vec n a) (:>) `f` y `f` ys)
 
 tVec :: DataType
 tVec = mkDataType "Vec" [cNil, cCons]
@@ -313,28 +324,48 @@ instance KnownNat n => Applicative (Vec n) where
 "zipWith$map" forall f xs ys. zipWith (\g a -> g a) (map f xs) ys = zipWith f xs ys
   #-}
 
-instance (KnownNat n, 1 <= n) => F.Foldable (Vec n) where
-  fold      = leToPlus @1 @n $ fold mappend
-  foldMap f = leToPlus @1 @n $ fold mappend . map f
+instance KnownNat n => F.Foldable (Vec n) where
+  fold Nil      = mempty
+  fold z@Cons{} = fold mappend z
+  foldMap _ Nil      = mempty
+  foldMap f z@Cons{} = fold mappend (map f z)
   foldr     = foldr
   foldl     = foldl
-  foldr1 f  = leToPlus @1 @n $ foldr1 f
-  foldl1 f  = leToPlus @1 @n $ foldl1 f
+  foldr1 _ Nil      = clashCompileError "foldr1: empty Vec"
+  foldr1 f z@Cons{} = foldr1 f z
+  foldl1 _ Nil      = clashCompileError "foldl1: empty Vec"
+  foldl1 f z@Cons{} = foldl1 f z
   toList    = toList
+  null Nil  = True
   null _    = False
   length    = length
-  maximum   = leToPlus @1 @n $ fold (\x y -> if x >= y then x else y)
-  minimum   = leToPlus @1 @n $ fold (\x y -> if x <= y then x else y)
-  sum       = leToPlus @1 @n $ fold (+)
-  product   = leToPlus @1 @n $ fold (*)
+  maximum Nil      = clashCompileError "maximum: empty Vec"
+  maximum z@Cons{} = fold (\x y -> if x >= y then x else y) z
+  minimum Nil      = clashCompileError "minimum: empty Vec"
+  minimum z@Cons{} = fold (\x y -> if x <= y then x else y) z
+  sum Nil      = 0
+  sum z@Cons{} = fold (+) z
+  product Nil      = 1
+  product z@Cons{} = fold (*) z
+
+#if MIN_VERSION_base(4,18,0)
+instance (KnownNat n, 1 <= n) => F1.Foldable1 (Vec n) where
+  fold1         = leToPlus @1 @n $ fold (<>)
+  foldMap1 f    = leToPlus @1 @n $ fold (<>) . map f
+  maximum       = leToPlus @1 @n maximum
+  minimum       = leToPlus @1 @n minimum
+  head          = leToPlus @1 @n head
+  last          = leToPlus @1 @n last
+#endif
 
 instance Functor (Vec n) where
   fmap = map
 
-instance (KnownNat n, 1 <= n) => Traversable (Vec n) where
+instance KnownNat n => Traversable (Vec n) where
   traverse = traverse#
 
-{-# NOINLINE traverse# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE traverse# #-}
 {-# ANN traverse# hasBlackBox #-}
 traverse# :: forall a f b n . Applicative f => (a -> f b) -> Vec n a -> f (Vec n b)
 traverse# _ Nil           = pure Nil
@@ -369,7 +400,8 @@ instance (NFDataX a, KnownNat n) => NFDataX (Vec n a) where
 singleton :: a -> Vec 1 a
 singleton = (`Cons` Nil)
 
-{-# NOINLINE head #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE head #-}
 {-# ANN head hasBlackBox #-}
 {- | Extract the first element of a vector
 
@@ -403,7 +435,8 @@ singleton = (`Cons` Nil)
 head :: Vec (n + 1) a -> a
 head (x `Cons` _) = x
 
-{-# NOINLINE tail #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE tail #-}
 {-# ANN tail hasBlackBox #-}
 {- | Extract the elements after the head of a vector
 
@@ -437,7 +470,8 @@ head (x `Cons` _) = x
 tail :: Vec (n + 1) a -> Vec n a
 tail (_ `Cons` xs) = xs
 
-{-# NOINLINE last #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE last #-}
 {-# ANN last hasBlackBox #-}
 {- | Extract the last element of a vector
 
@@ -472,7 +506,8 @@ last :: Vec (n + 1) a -> a
 last (x `Cons` Nil)         = x
 last (_ `Cons` y `Cons` ys) = last (y `Cons` ys)
 
-{-# NOINLINE init #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE init #-}
 {-# ANN init hasBlackBox #-}
 {- | Extract all the elements of a vector except the last element
 
@@ -640,7 +675,8 @@ infixr 5 ++
 (++) :: Vec n a -> Vec m a -> Vec (n + m) a
 Nil           ++ ys = ys
 (x `Cons` xs) ++ ys = x `Cons` xs ++ ys
-{-# NOINLINE (++) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE (++) #-}
 {-# ANN (++) hasBlackBox #-}
 
 -- | Split a vector into two vectors at the given point.
@@ -651,7 +687,8 @@ Nil           ++ ys = ys
 -- (1 :> 2 :> 3 :> Nil,7 :> 8 :> Nil)
 splitAt :: SNat m -> Vec (m + n) a -> (Vec m a, Vec n a)
 splitAt n xs = splitAtU (toUNat n) xs
-{-# NOINLINE splitAt #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE splitAt #-}
 {-# ANN splitAt hasBlackBox #-}
 
 splitAtU :: UNat m -> Vec (m + n) a -> (Vec m a, Vec n a)
@@ -675,7 +712,8 @@ splitAtI = withSNat splitAt
 concat :: Vec n (Vec m a) -> Vec (n * m) a
 concat Nil           = Nil
 concat (x `Cons` xs) = x ++ concat xs
-{-# NOINLINE concat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE concat #-}
 {-# ANN concat hasBlackBox #-}
 
 -- | Map a function over all the elements of a vector and concatentate the resulting vectors.
@@ -693,7 +731,8 @@ concatMap f xs = concat (map f xs)
 -- (1 :> 2 :> 3 :> 4 :> Nil) :> (5 :> 6 :> 7 :> 8 :> Nil) :> (9 :> 10 :> 11 :> 12 :> Nil) :> Nil
 unconcat :: KnownNat n => SNat m -> Vec (n * m) a -> Vec n (Vec m a)
 unconcat n xs = unconcatU (withSNat toUNat) (toUNat n) xs
-{-# NOINLINE unconcat #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unconcat #-}
 {-# ANN unconcat hasBlackBox #-}
 
 unconcatU :: UNat n -> UNat m -> Vec (n * m) a -> Vec n (Vec m a)
@@ -725,7 +764,8 @@ merge x y = concat (transpose (x :> singleton y))
 reverse :: Vec n a -> Vec n a
 reverse Nil           = Nil
 reverse (x `Cons` xs) = reverse xs :< x
-{-# NOINLINE reverse #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE reverse #-}
 {-# ANN reverse hasBlackBox #-}
 
 -- | \"'map' @f xs@\" is the vector obtained by applying /f/ to each element
@@ -739,7 +779,8 @@ reverse (x `Cons` xs) = reverse xs :< x
 map :: (a -> b) -> Vec n a -> Vec n b
 map _ Nil           = Nil
 map f (x `Cons` xs) = f x `Cons` map f xs
-{-# NOINLINE map #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE map #-}
 {-# ANN map hasBlackBox #-}
 
 -- | Apply a function of every element of a vector and its index.
@@ -749,7 +790,7 @@ map f (x `Cons` xs) = f x `Cons` map f xs
 -- >>> imap (+) (2 :> 2 :> 2 :> 2 :> Nil)
 -- 2 :> 3 :> *** Exception: X: Clash.Sized.Index: result 4 is out of bounds: [0..3]
 -- ...
--- >>> imap (\i a -> fromIntegral i + a) (2 :> 2 :> 2 :> 2 :> Nil) :: Vec 4 (Unsigned 8)
+-- >>> imap (\i a -> extend (bitCoerce i) + a) (2 :> 2 :> 2 :> 2 :> Nil) :: Vec 4 (Unsigned 8)
 -- 2 :> 3 :> 4 :> 5 :> Nil
 --
 -- \"'imap' @f xs@\" corresponds to the following circuit layout:
@@ -762,12 +803,13 @@ imap f = go 0
     go :: Index n -> Vec m a -> Vec m b
     go _ Nil           = Nil
     go n (x `Cons` xs) = f n x `Cons` go (n+1) xs
-{-# NOINLINE imap #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE imap #-}
 {-# ANN imap hasBlackBox #-}
 
 {- | Zip two vectors with a functions that also takes the elements' indices.
 
-#if __GLASGOW_HASKELL__ >= 900
+#if __GLASGOW_HASKELL__ >= 900 && __GLASGOW_HASKELL__ < 904
 >>> izipWith (\i a b -> i + a + b) (2 :> 2 :> Nil)  (3 :> 3:> Nil)
 *** Exception: X: Clash.Sized.Index: result 2 is out of bounds: [0..1]
 ...
@@ -776,14 +818,14 @@ imap f = go 0
 *** Exception: X: Clash.Sized.Index: result 3 is out of bounds: [0..1]
 ...
 #endif
->>> izipWith (\i a b -> fromIntegral i + a + b) (2 :> 2 :> Nil) (3 :> 3 :> Nil) :: Vec 2 (Unsigned 8)
+>>> izipWith (\i a b -> extend (bitCoerce i) + a + b) (2 :> 2 :> Nil) (3 :> 3 :> Nil) :: Vec 2 (Unsigned 8)
 5 :> 6 :> Nil
 
 \"'imap' @f xs@\" corresponds to the following circuit layout:
 
 <<doc/izipWith.svg>>
 
-__NB:__ 'izipWith' is /strict/ in its second argument, and /lazy/ in its
+__NB__: 'izipWith' is /strict/ in its second argument, and /lazy/ in its
 third. This matters when 'izipWith' is used in a recursive setting. See
 'lazyV' for more information.
 -}
@@ -877,13 +919,14 @@ elemIndex x = findIndex (x ==)
 --
 -- <<doc/zipWith.svg>>
 --
--- __NB:__ 'zipWith' is /strict/ in its second argument, and /lazy/ in its
+-- __NB__: 'zipWith' is /strict/ in its second argument, and /lazy/ in its
 -- third. This matters when 'zipWith' is used in a recursive setting. See
 -- 'lazyV' for more information.
 zipWith :: (a -> b -> c) -> Vec n a -> Vec n b -> Vec n c
 zipWith _ Nil           _  = Nil
 zipWith f (x `Cons` xs) ys = f x (head ys) `Cons` zipWith f xs (tail ys)
-{-# NOINLINE zipWith #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE zipWith #-}
 {-# ANN zipWith hasBlackBox #-}
 
 -- | 'zipWith3' generalizes 'zip3' by zipping with the function given
@@ -895,7 +938,7 @@ zipWith f (x `Cons` xs) ys = f x (head ys) `Cons` zipWith f xs (tail ys)
 --
 -- <<doc/zipWith3.svg>>
 --
--- __NB:__ 'zipWith3' is /strict/ in its second argument, and /lazy/ in its
+-- __NB__: 'zipWith3' is /strict/ in its second argument, and /lazy/ in its
 -- third and fourth. This matters when 'zipWith3' is used in a recursive setting.
 -- See 'lazyV' for more information.
 zipWith3 :: (a -> b -> c -> d) -> Vec n a -> Vec n b -> Vec n c -> Vec n d
@@ -904,7 +947,7 @@ zipWith3 f us vs ws = zipWith (\a (b,c) -> f a b c) us (zip vs ws)
 
 -- 'zipWith4' is analogous to 'zipWith3', but with four vectors.
 --
--- __NB:__ 'zipWith4' is /strict/ in its second argument, and /lazy/ its following
+-- __NB__: 'zipWith4' is /strict/ in its second argument, and /lazy/ its following
 -- arguments. This matters when 'zipWith4' is used in a recursive setting. See
 -- 'lazyV' for more information.
 zipWith4
@@ -920,7 +963,7 @@ zipWith4 f us vs ws xs =
 
 -- 'zipWith5' is analogous to 'zipWith3', but with five vectors.
 --
--- __NB:__ 'zipWith5' is /strict/ in its second argument, and /lazy/ its following
+-- __NB__: 'zipWith5' is /strict/ in its second argument, and /lazy/ its following
 -- arguments. This matters when 'zipWith5' is used in a recursive setting. See
 -- 'lazyV' for more information.
 zipWith5
@@ -937,7 +980,7 @@ zipWith5 f us vs ws xs ys =
 
 -- 'zipWith6' is analogous to 'zipWith3', but with six vectors.
 --
--- __NB:__ 'zipWith6' is /strict/ in its second argument, and /lazy/ its following
+-- __NB__: 'zipWith6' is /strict/ in its second argument, and /lazy/ its following
 -- arguments. This matters when 'zipWith6' is used in a recursive setting. See
 -- 'lazyV' for more information.
 zipWith6
@@ -955,7 +998,7 @@ zipWith6 f us vs ws xs ys zs =
 
 -- 'zipWith7' is analogous to 'zipWith3', but with seven vectors.
 --
--- __NB:__ 'zipWith7' is /strict/ in its second argument, and /lazy/ its following
+-- __NB__: 'zipWith7' is /strict/ in its second argument, and /lazy/ its following
 -- arguments. This matters when 'zipWith7' is used in a recursive setting. See
 -- 'lazyV' for more information.
 zipWith7
@@ -993,7 +1036,8 @@ zipWith7 f ts us vs ws xs ys zs =
 foldr :: (a -> b -> b) -> b -> Vec n a -> b
 foldr _ z Nil           = z
 foldr f z (x `Cons` xs) = f x (foldr f z xs)
-{-# NOINLINE foldr #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE foldr #-}
 {-# ANN foldr hasBlackBox #-}
 
 -- | 'foldl', applied to a binary operator, a starting value (typically
@@ -1087,8 +1131,13 @@ fold f vs = fold' (toList vs)
     fold' xs  = fold' ys `f` fold' zs
       where
         (ys,zs) = P.splitAt (P.length xs `div` 2) xs
-{-# NOINLINE fold #-}
-{-# ANN fold (InlinePrimitive [VHDL,Verilog,SystemVerilog] "[ { \"BlackBoxHaskell\" : { \"name\" : \"Clash.Sized.Vector.fold\", \"templateFunction\" : \"Clash.Primitives.Sized.Vector.foldBBF\"}} ]") #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fold #-}
+{-# ANN fold (InlineYamlPrimitive [VHDL,Verilog,SystemVerilog] [I.__i|
+  BlackBoxHaskell:
+    name: Clash.Sized.Vector.fold
+    templateFunction: Clash.Primitives.Sized.Vector.foldBBF
+  |]) #-}
 
 -- | 'scanl' is similar to 'foldl', but returns a vector of successive reduced
 -- values from the left:
@@ -1114,6 +1163,22 @@ scanl f z xs = ws
     ws = z `Cons` zipWith (flip f) xs (init ws)
 {-# INLINE scanl #-}
 
+-- | 'scanl' with no seed value
+--
+-- >>> scanl1 (-) (1 :> 2 :> 3 :> 4 :> Nil)
+-- 1 :> -1 :> -4 :> -8 :> Nil
+scanl1 :: KnownNat n => (a -> a -> a) -> Vec (n+1) a -> Vec (n+1) a
+scanl1 op (v:>vs) = scanl op v vs
+{-# INLINE scanl1 #-}
+
+-- | 'scanr' with no seed value
+--
+-- >>> scanr1 (-) (1 :> 2 :> 3 :> 4 :> Nil)
+-- -2 :> 3 :> -1 :> 4 :> Nil
+scanr1 :: KnownNat n => (a -> a -> a) -> Vec (n+1) a -> Vec (n+1) a
+scanr1 op vs = scanr op (last vs) (init vs)
+{-# INLINE scanr1 #-}
+
 -- | 'postscanl' is a variant of 'scanl' where the first result is dropped:
 --
 -- > postscanl f z (x1 :> x2 :> ... :> Nil) == (z `f` x1) :> ((z `f` x1) `f` x2) :> ... :> Nil
@@ -1349,12 +1414,13 @@ index_int xs i@(I# n0)
     sub (y `Cons` (!ys)) n = if isTrue# (n ==# 0#)
                                 then y
                                 else sub ys (n -# 1#)
-{-# NOINLINE index_int #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE index_int #-}
 {-# ANN index_int hasBlackBox #-}
 
 -- | \"@xs@ '!!' @n@\" returns the /n/'th element of /xs/.
 --
--- __NB__: vector elements have an __ASCENDING__ subscript starting from 0 and
+-- __NB__: Vector elements have an __ASCENDING__ subscript starting from 0 and
 -- ending at @'length' - 1@.
 --
 -- >>> (1:>2:>3:>4:>5:>Nil) !! 4
@@ -1376,7 +1442,8 @@ xs !! i = index_int xs (fromEnum i)
 -- 3
 length :: KnownNat n => Vec n a -> Int
 length = fromInteger . natVal . asNatProxy
-{-# NOINLINE length #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE length #-}
 {-# ANN length hasBlackBox #-}
 
 replace_int :: KnownNat n => Vec n a -> Int -> a -> Vec n a
@@ -1393,13 +1460,14 @@ replace_int xs i@(I# n0) a
     sub (y `Cons` (!ys)) n b = if isTrue# (n ==# 0#)
                                  then b `Cons` ys
                                  else y `Cons` sub ys (n -# 1#) b
-{-# NOINLINE replace_int #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE replace_int #-}
 {-# ANN replace_int hasBlackBox #-}
 
 -- | \"'replace' @n a xs@\" returns the vector /xs/ where the /n/'th element is
 -- replaced by /a/.
 --
--- __NB__: vector elements have an __ASCENDING__ subscript starting from 0 and
+-- __NB__: Vector elements have an __ASCENDING__ subscript starting from 0 and
 -- ending at @'length' - 1@.
 --
 -- >>> replace 3 7 (1:>2:>3:>4:>5:>Nil)
@@ -1422,7 +1490,19 @@ replace i y xs = replace_int xs (fromEnum i) y
 >>> take d0               (1:>2:>Nil)
 Nil
 
-#if __GLASGOW_HASKELL__ >= 900
+#if __GLASGOW_HASKELL__ == 906
+>>> take d4               (1:>2:>Nil)
+<BLANKLINE>
+<interactive>:...
+    • Couldn't match type ‘4 + n0’ with ‘2’
+      Expected: Vec (4 + n0) a
+        Actual: Vec (1 + 1) a
+        The type variable ‘n0’ is ambiguous
+    • In the second argument of ‘take’, namely ‘(1 :> 2 :> Nil)’
+      In the expression: take d4 (1 :> 2 :> Nil)
+      In an equation for ‘it’: it = take d4 (1 :> 2 :> Nil)
+
+#elif __GLASGOW_HASKELL__ >= 900
 >>> take d4               (1:>2:>Nil)
 <BLANKLINE>
 <interactive>:...
@@ -1460,21 +1540,35 @@ takeI :: KnownNat m => Vec (m + n) a -> Vec m a
 takeI = withSNat take
 {-# INLINE takeI #-}
 
--- | \"'drop' @n xs@\" returns the suffix of /xs/ after the first /n/ elements.
---
--- >>> drop (SNat :: SNat 3) (1:>2:>3:>4:>5:>Nil)
--- 4 :> 5 :> Nil
--- >>> drop d3               (1:>2:>3:>4:>5:>Nil)
--- 4 :> 5 :> Nil
--- >>> drop d0               (1:>2:>Nil)
--- 1 :> 2 :> Nil
--- >>> drop d4               (1:>2:>Nil)
--- <BLANKLINE>
--- <interactive>:...: error:
---     • Couldn't match...type ‘4 + n0...
---       The type variable ‘n0’ is ambiguous
---     • In the first argument of ‘print’, namely ‘it’
---       In a stmt of an interactive GHCi command: print it
+{- | \"'drop' @n xs@\" returns the suffix of /xs/ after the first /n/ elements.
+
+>>> drop (SNat :: SNat 3) (1:>2:>3:>4:>5:>Nil)
+4 :> 5 :> Nil
+>>> drop d3               (1:>2:>3:>4:>5:>Nil)
+4 :> 5 :> Nil
+>>> drop d0               (1:>2:>Nil)
+1 :> 2 :> Nil
+
+#if __GLASGOW_HASKELL__ == 906
+>>> drop d4               (1:>2:>Nil)
+<BLANKLINE>
+<interactive>:...: error:...
+    • Couldn't match...type ‘4 + n0...
+        The type variable ‘n0’ is ambiguous
+    • In the first argument of ‘print’, namely ‘it’
+      In a stmt of an interactive GHCi command: print it
+
+#else
+>>> drop d4               (1:>2:>Nil)
+<BLANKLINE>
+<interactive>:...: error:...
+    • Couldn't match...type ‘4 + n0...
+      The type variable ‘n0’ is ambiguous
+    • In the first argument of ‘print’, namely ‘it’
+      In a stmt of an interactive GHCi command: print it
+
+#endif
+-}
 drop :: SNat m -> Vec (m + n) a -> Vec n a
 drop n = snd . splitAt n
 {-# INLINE drop #-}
@@ -1489,7 +1583,7 @@ dropI = withSNat drop
 
 -- | \"'at' @n xs@\" returns /n/'th element of /xs/
 --
--- __NB__: vector elements have an __ASCENDING__ subscript starting from 0 and
+-- __NB__: Vector elements have an __ASCENDING__ subscript starting from 0 and
 -- ending at @'length' - 1@.
 --
 -- >>> at (SNat :: SNat 1) (1:>2:>3:>4:>5:>Nil)
@@ -1519,7 +1613,8 @@ select f s n xs = select' (toUNat n) $ drop f xs
     select' UZero      _               = Nil
     select' (USucc n') vs@(x `Cons` _) = x `Cons`
                                          select' n' (drop s (unsafeCoerce vs))
-{-# NOINLINE select #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE select #-}
 {-# ANN select hasBlackBox #-}
 
 -- | \"'selectI' @f s xs@\" selects as many elements as demanded by the context
@@ -1543,7 +1638,8 @@ selectI f s xs = withSNat (\n -> select f s n xs)
 -- 6 :> 6 :> 6 :> Nil
 replicate :: SNat n -> a -> Vec n a
 replicate n a = replicateU (toUNat n) a
-{-# NOINLINE replicate #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE replicate #-}
 {-# ANN replicate hasBlackBox #-}
 
 replicateU :: UNat n -> a -> Vec n a
@@ -1575,7 +1671,7 @@ iterate :: SNat n -> (a -> a) -> a -> Vec n a
 iterate SNat = iterateI
 {-# INLINE iterate #-}
 
--- | \"'iterate' @f x@\" returns a vector starting with @x@ followed by @n@
+-- | \"'iterateI' @f x@\" returns a vector starting with @x@ followed by @n@
 -- repeated applications of @f@ to @x@, where @n@ is determined by the context.
 --
 -- > iterateI f x :: Vec 3 a == (x :> f x :> f (f x) :> Nil)
@@ -1591,14 +1687,15 @@ iterateI f a = xs
   where
     xs = init (a `Cons` ws)
     ws = map f (lazyV xs)
-{-# NOINLINE iterateI #-}
-{-# ANN iterateI (InlinePrimitive [VHDL,Verilog,SystemVerilog] [I.i| [{
-    "BlackBoxHaskell": {
-        "name": "Clash.Sized.Vector.iterateI"
-      , "templateFunction": "Clash.Primitives.Sized.Vector.iterateBBF"
-    }}] |]) #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE iterateI #-}
+{-# ANN iterateI (InlineYamlPrimitive [VHDL,Verilog,SystemVerilog] [I.__i|
+  BlackBoxHaskell:
+    name: Clash.Sized.Vector.iterateI
+    templateFunction: Clash.Primitives.Sized.Vector.iterateBBF
+  |]) #-}
 
--- | \"'unfoldr @n f s@\" builds a vector of length @n@ from a seed value @s@,
+-- | \"'unfoldr' @n f s@\" builds a vector of length @n@ from a seed value @s@,
 -- where every element @a@ is created by successive calls of @f@ on @s@. Unlike
 -- 'Data.List.unfoldr' from "Data.List" the generating function @f@ cannot
 -- dictate the length of the resulting vector, it must be statically known.
@@ -1611,7 +1708,7 @@ unfoldr :: SNat n -> (s -> (a,s)) -> s -> Vec n a
 unfoldr SNat = unfoldrI
 {-# INLINE unfoldr #-}
 
--- | \"'unfoldr @f s@\" builds a vector from a seed value @s@, where every
+-- | \"'unfoldrI' @f s@\" builds a vector from a seed value @s@, where every
 -- element @a@ is created by successive calls of @f@ on @s@; the length of the
 -- vector is inferred from the context. Unlike 'Data.List.unfoldr' from
 -- "Data.List" the generating function @f@ cannot  dictate the length of the
@@ -1668,7 +1765,8 @@ generateI f a = iterateI f (f a)
 -- (1 :> 3 :> 5 :> Nil) :> (2 :> 4 :> 6 :> Nil) :> Nil
 transpose :: KnownNat n => Vec m (Vec n a) -> Vec n (Vec m a)
 transpose = traverse# id
-{-# NOINLINE transpose #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE transpose #-}
 {-# ANN transpose hasBlackBox #-}
 
 -- | 1-dimensional stencil computations
@@ -1692,20 +1790,29 @@ stencil1d :: KnownNat n
 stencil1d stX f xs = map f (windows1d stX xs)
 {-# INLINE stencil1d #-}
 
--- | 2-dimensional stencil computations
---
--- \"'stencil2d' @stY stX f xss@\", where /xss/ is a matrix of /stY + m/ rows
--- of /stX + n/ elements, applies the stencil computation /f/ on:
--- /(m + 1) * (n + 1)/ overlapping (2D) windows of /stY/ rows of /stX/ elements,
--- drawn from /xss/. The result matrix has /m + 1/ rows of /n + 1/ elements.
---
--- >>> let xss = ((1:>2:>3:>4:>Nil):>(5:>6:>7:>8:>Nil):>(9:>10:>11:>12:>Nil):>(13:>14:>15:>16:>Nil):>Nil)
--- >>> :t xss
--- xss :: Num a => Vec 4 (Vec 4 a)
--- >>> :t stencil2d d2 d2 (sum . map sum) xss
--- stencil2d d2 d2 (sum . map sum) xss :: Num b => Vec 3 (Vec 3 b)
--- >>> stencil2d d2 d2 (sum . map sum) xss
--- (14 :> 18 :> 22 :> Nil) :> (30 :> 34 :> 38 :> Nil) :> (46 :> 50 :> 54 :> Nil) :> Nil
+{- | 2-dimensional stencil computations
+
+\"'stencil2d' @stY stX f xss@\", where /xss/ is a matrix of /stY + m/ rows
+of /stX + n/ elements, applies the stencil computation /f/ on:
+/(m + 1) * (n + 1)/ overlapping (2D) windows of /stY/ rows of /stX/ elements,
+drawn from /xss/. The result matrix has /m + 1/ rows of /n + 1/ elements.
+
+>>> let xss = ((1:>2:>3:>4:>Nil):>(5:>6:>7:>8:>Nil):>(9:>10:>11:>12:>Nil):>(13:>14:>15:>16:>Nil):>Nil)
+>>> :t xss
+xss :: Num a => Vec 4 (Vec 4 a)
+
+#if __GLASGOW_HASKELL__ >= 902
+>>> :t stencil2d d2 d2 (sum . map sum) xss
+stencil2d d2 d2 (sum . map sum) xss :: Num a => Vec 3 (Vec 3 a)
+
+#else
+>>> :t stencil2d d2 d2 (sum . map sum) xss
+stencil2d d2 d2 (sum . map sum) xss :: Num b => Vec 3 (Vec 3 b)
+
+#endif
+>>> stencil2d d2 d2 (sum . map sum) xss
+(14 :> 18 :> 22 :> Nil) :> (30 :> 34 :> 38 :> Nil) :> (46 :> 50 :> 54 :> Nil) :> Nil
+-}
 stencil2d :: (KnownNat n, KnownNat m)
           => SNat (stY + 1) -- ^ Window hight /stY/, at least size 1
           -> SNat (stX + 1) -- ^ Window width /stX/, at least size 1
@@ -1860,7 +1967,7 @@ interleave d = concat . transpose . unconcat d
 -- >>> rotateLeft xs (-1)
 -- 4 :> 1 :> 2 :> 3 :> Nil
 --
--- __NB:__ use `rotateLeftS` if you want to rotate left by a /static/ amount.
+-- __NB__: Use `rotateLeftS` if you want to rotate left by a /static/ amount.
 rotateLeft :: (Enum i, KnownNat n)
            => Vec n a
            -> i
@@ -1881,7 +1988,7 @@ rotateLeft xs i = map ((xs !!) . (`mod` len)) (iterateI (+1) i')
 -- >>> rotateRight xs (-1)
 -- 2 :> 3 :> 4 :> 1 :> Nil
 --
--- __NB:__ use `rotateRightS` if you want to rotate right by a /static/ amount.
+-- __NB__: Use `rotateRightS` if you want to rotate right by a /static/ amount.
 rotateRight :: (Enum i, KnownNat n)
             => Vec n a
             -> i
@@ -1898,7 +2005,7 @@ rotateRight xs i = map ((xs !!) . (`mod` len)) (iterateI (+1) i')
 -- >>> rotateLeftS xs d1
 -- 2 :> 3 :> 4 :> 1 :> Nil
 --
--- __NB:__ use `rotateLeft` if you want to rotate left by a /dynamic/ amount.
+-- __NB__: Use `rotateLeft` if you want to rotate left by a /dynamic/ amount.
 rotateLeftS :: KnownNat n
             => Vec n a
             -> SNat d
@@ -1909,7 +2016,8 @@ rotateLeftS xs d = go (snatToInteger d `mod` natVal (asNatProxy xs)) xs
     go _ Nil           = Nil
     go 0 ys            = ys
     go n (y `Cons` ys) = go (n-1) (ys :< y)
-{-# NOINLINE rotateLeftS #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rotateLeftS #-}
 {-# ANN rotateLeftS hasBlackBox #-}
 
 -- | /Statically/ rotate a 'Vec'tor to the right:
@@ -1918,7 +2026,7 @@ rotateLeftS xs d = go (snatToInteger d `mod` natVal (asNatProxy xs)) xs
 -- >>> rotateRightS xs d1
 -- 4 :> 1 :> 2 :> 3 :> Nil
 --
--- __NB:__ use `rotateRight` if you want to rotate right by a /dynamic/ amount.
+-- __NB__: Use `rotateRight` if you want to rotate right by a /dynamic/ amount.
 rotateRightS :: KnownNat n
              => Vec n a
              -> SNat d
@@ -1928,7 +2036,8 @@ rotateRightS xs d = go (snatToInteger d `mod` natVal (asNatProxy xs)) xs
     go _ Nil            = Nil
     go 0 ys             = ys
     go n ys@(Cons _ _)  = go (n-1) (last ys :> init ys)
-{-# NOINLINE rotateRightS #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE rotateRightS #-}
 {-# ANN rotateRightS hasBlackBox #-}
 
 -- | Convert a vector to a list.
@@ -1936,7 +2045,7 @@ rotateRightS xs d = go (snatToInteger d `mod` natVal (asNatProxy xs)) xs
 -- >>> toList (1:>2:>3:>Nil)
 -- [1,2,3]
 --
--- __NB:__ this function is not synthesizable
+-- __NB__: This function is not synthesizable
 toList :: Vec n a -> [a]
 toList = foldr (:) []
 {-# INLINE toList #-}
@@ -1953,8 +2062,8 @@ toList = foldr (:) []
 -- >>> Vec.fromList [1,2,3,4,5] :: Maybe (Vec 10 Int)
 -- Nothing
 --
--- * __NB:__ use `listToVecTH` if you want to make a /statically known/ vector
--- * __NB:__ this function is not synthesizable
+-- * __NB__: Use `listToVecTH` if you want to make a /statically known/ vector
+-- * __NB__: This function is not synthesizable
 --
 fromList :: forall n a. (KnownNat n) => [a] -> Maybe (Vec n a)
 fromList xs
@@ -1964,7 +2073,8 @@ fromList xs
   exactLength 0 acc = null acc
   exactLength _ []  = False
   exactLength i (_:ys) = exactLength (i - 1) ys
-{-# NOINLINE fromList #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE fromList #-}
 {-# ANN fromList dontTranslate #-}
 
 -- | Convert a list to a vector. This function always returns a vector of the
@@ -1981,8 +2091,8 @@ fromList xs
 -- 1 :> 2 :> 3 :> 4 :> 5 :> *** Exception: Clash.Sized.Vector.unsafeFromList: vector larger than list
 -- ...
 --
--- * __NB:__ use `listToVecTH` if you want to make a /statically known/ vector
--- * __NB:__ this function is not synthesizable
+-- * __NB__: Use `listToVecTH` if you want to make a /statically known/ vector
+-- * __NB__: This function is not synthesizable
 --
 unsafeFromList :: forall n a. (KnownNat n) => [a] -> Vec n a
 unsafeFromList = unfoldr SNat go
@@ -1992,7 +2102,8 @@ unsafeFromList = unfoldr SNat go
   go [] =
     let item = error "Clash.Sized.Vector.unsafeFromList: vector larger than list"
      in (item, [])
-{-# NOINLINE unsafeFromList #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unsafeFromList #-}
 {-# ANN unsafeFromList dontTranslate #-}
 
 -- | Create a vector literal from a list literal.
@@ -2093,7 +2204,8 @@ lazyV = lazyV' (repeat ())
     lazyV' :: Vec n () -> Vec n a -> Vec n a
     lazyV' Nil           _  = Nil
     lazyV' (_ `Cons` xs) ys = head ys `Cons` lazyV' xs (tail ys)
-{-# NOINLINE lazyV #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE lazyV #-}
 {-# ANN lazyV hasBlackBox #-}
 
 -- | A /dependently/ typed fold.
@@ -2197,7 +2309,8 @@ dfold _ f z xs = go (snatProxy (asNatProxy xs)) xs
     go s (y `Cons` ys) =
       let s' = s `subSNat` d1
       in  f s' y (go s' ys)
-{-# NOINLINE dfold #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE dfold #-}
 {-# ANN dfold hasBlackBox #-}
 
 {- | A combination of 'dfold' and 'fold': a /dependently/ typed fold that
@@ -2360,7 +2473,8 @@ dtfold _ f g = go (SNat :: SNat k)
           sn'       = sn `subSNat` d1
           (xsL,xsR) = splitAt (pow2SNat sn') xs
       in  g sn' (go sn' xsL) (go sn' xsR)
-{-# NOINLINE dtfold #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE dtfold #-}
 {-# ANN dtfold hasBlackBox #-}
 
 -- | To be used as the motive /p/ for 'dfold', when the /f/ in \"'dfold' @p f@\"
@@ -2404,6 +2518,20 @@ vfold :: forall k a b . KnownNat k
 vfold f xs = dfold (Proxy @(VCons b)) f Nil xs
 {-# INLINE vfold #-}
 
+-- | The largest element of a non-empty vector
+maximum ::
+  Ord a =>
+  Vec (n + 1) a ->
+  a
+maximum = fold (\x y -> if x >= y then x else y)
+
+-- | The least element of a non-empty vector
+minimum ::
+  Ord a =>
+  Vec (n + 1) a ->
+  a
+minimum = fold (\x y -> if x <= y then x else y)
+
 -- | Apply a function to every element of a vector and the element's position
 -- (as an 'SNat' value) in the vector.
 --
@@ -2437,7 +2565,8 @@ concatBitVector# = go 0
   go (BV accMsk accVal) ((BV xMsk xVal) `Cons` xs) =
     let sh = fromInteger (natVal (Proxy @m)) :: Int in
     go (BV (shiftL accMsk sh .|. xMsk) (shiftL accVal sh .|. xVal)) xs
-{-# NOINLINE concatBitVector# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE concatBitVector# #-}
 {-# ANN concatBitVector# hasBlackBox #-}
 
 unconcatBitVector#
@@ -2453,7 +2582,8 @@ unconcatBitVector# orig = snd (go (toUNat (SNat @n)))
       let (bv,xs) = go n
           (l,x) = (GHC.Magic.noinline split#) bv
       in  (l,x :> xs)
-{-# NOINLINE unconcatBitVector# #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE unconcatBitVector# #-}
 {-# ANN unconcatBitVector# hasBlackBox #-}
 
 -- | Convert a 'BitVector' to a 'Vec' of 'Bit's.
@@ -2485,7 +2615,8 @@ seqV
 seqV v b =
   let s () e = seq e () in
   foldl s () v `seq` b
-{-# NOINLINE seqV #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE seqV #-}
 {-# ANN seqV hasBlackBox #-}
 infixr 0 `seqV`
 
@@ -2508,7 +2639,8 @@ seqVX
 seqVX v b =
   let s () e = seqX e () in
   foldl s () v `seqX` b
-{-# NOINLINE seqVX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE seqVX #-}
 {-# ANN seqVX hasBlackBox #-}
 infixr 0 `seqVX`
 
