@@ -2,8 +2,9 @@
 Copyright : © 2014-2016, Christiaan Baaij,
               2017-2019, Myrtle Software Ltd
               2017     , QBayLogic, Google Inc.,
-              2021     , QBayLogic B.V.
+              2021-2023, QBayLogic B.V.
               2022     , LUMI GUIDE FIETSDETECTIE B.V.
+              2023     , Alex Mason
 
 Licence   : Creative Commons 4.0 (CC BY 4.0) (https://creativecommons.org/licenses/by/4.0/)
 Maintainer:  QBayLogic B.V. <devops@qbaylogic.com>
@@ -80,6 +81,7 @@ module Clash.Tutorial (
 where
 
 import Clash.Prelude
+import Clash.Prelude.Mealy
 import Clash.Explicit.Testbench
 import Clash.XException (NFDataX)
 import Control.Monad.ST
@@ -96,6 +98,7 @@ import Data.Default.Class
 >>> :set -XTemplateHaskell -XDataKinds -XConstraintKinds -XTypeApplications
 >>> :m -Prelude
 >>> import Clash.Prelude
+>>> import Clash.Prelude.Mealy
 >>> import Clash.Explicit.Testbench
 >>> :{
 let ma :: Num a => a -> (a, a) -> a
@@ -232,15 +235,22 @@ at the same time. If you followed the installation instructions based on
 start the Clash compiler in interpretive mode by:
 
 @
-stack exec --package clash-ghc -- clashi
+stack exec --resolver lts-19 --package clash-ghc -- clashi
 @
 
-If you followed the installation instruction based on
-<https://snapcraft.io/clash snap> (Linux only), you can start the Clash compiler
-in interpretive mode by:
+If instead you followed the [instructions](https://clash-lang.org/install/linux/)
+to setup a starter project with Stack, you can also run @clashi@ inside such a
+project. Change to the directory of the project, and invoke
 
 @
-clash.clashi
+stack run -- clashi
+@
+
+If you instead set up the starter project with GHC and Cabal, change to the
+directory of the project and invoke
+
+@
+cabal run -- clashi
 @
 
 For those familiar with Haskell/GHC, this is indeed just @GHCi@, with three
@@ -263,7 +273,7 @@ choice in editor, the following @edit-load-run@ cycle probably work best for you
       * Once a file has been edited and saved, type @:r@ to reload the files in
         the interpreter
 
-You are of course free to deviate from these suggestions as you see fit :-) It
+You are of course free to deviate from these suggestions as you see fit :-). It
 is just recommended that you have the Clash interpreter open during this
 tutorial.
 -}
@@ -329,8 +339,8 @@ You should read this as follows:
 
  * __@ma ::@__, @ma@ is of type..
 
- * __@Num a@__, there is some type called @a@ that is a @'Num'@. Examples of
-   instances of @'Num'@ are @'Int'@, @'Signed' 16@, @'Index' 32@, and @'Float'@.
+ * __@Num a@__, there is some type called @a@ that is a 'Num'. Examples of
+   instances of 'Num' are 'Int', @'Signed' 16@, @'Index' 32@, and 'Float'.
 
  * __@a@__, @ma@'s first argument is of type @a@
 
@@ -339,8 +349,8 @@ You should read this as follows:
  * __@a@__, @ma@'s result is of type @a@
 
 Note that @ma@ therefore works on multiple types! The only condition we
-imposed is that @a@ should be a @'Num'@ber type. In Clash this means it should
-support the operations @'Prelude.+'@, @'Prelude.-'@, @'Prelude.*'@, and some
+imposed is that @a@ should be a 'Num'ber type. In Clash this means it should
+support the operations 'Prelude.+', 'Prelude.-', 'Prelude.*', and some
 others. Indeed, this is why Clash adds the constraint in the first place: the
 definition of @ma@ uses @+@ and @*@. Whenever a function works over multiple
 types, we call it /polymorphic/ ("poly" meaning "many", "morphic" meaning
@@ -352,12 +362,12 @@ Talking about /types/ also brings us to one of the most important parts of this
 tutorial: /types/ and /synchronous sequential logic/. Especially how we can
 always determine, through the types of a specification, if it describes
 combinational logic or (synchronous) sequential logic. We do this by examining
-the definition of one of the sequential primitives, the @'register'@ function:
+the definition of one of the sequential primitives, the 'register' function:
 
 @
 register
-     ( 'HiddenClockResetEnable' dom
-     , 'Clash.XException.NFDataX' a )
+  :: ( 'HiddenClockResetEnable' dom
+     , 'NFDataX' a )
   => a
   -> 'Signal' dom a
   -> 'Signal' dom a
@@ -391,10 +401,10 @@ examine the 'register' function by taking a look at the first 4 samples of the
 Where we see that the initial value of the signal is the specified 0 value,
 followed by 8's. You might be surprised to see /two/ zeros instead of just a
 single zero. What happens is that in Clash you get to see the output of the
-circuit /before/ the clock becomes actives. In other words, in Clash you get to
+circuit /before/ the clock becomes active. In other words, in Clash you get to
 describe the powerup values of registers too. Whether this is a defined or
 unknown value depends on your hardware target, and can be configured by using a
-different synthesis @'Domain'@. The default synthesis domain, @'System', assumes
+different synthesis 'Domain'. The default synthesis domain, @'System', assumes
 that registers do have a powerup value - as is true for most FPGA platforms in
 most contexts.
 -}
@@ -445,7 +455,7 @@ shape of @macT@:
 
 @
 mealy
-  :: ('HiddenClockResetEnable' dom, 'Clash.XException.NFDataX' s)
+  :: ('HiddenClockResetEnable' dom, 'NFDataX' s)
   => (s -> i -> (s,o))
   -> s
   -> ('Signal' dom i -> 'Signal' dom o)
@@ -458,7 +468,7 @@ The complete sequential MAC circuit can now be specified as:
 mac inp = 'mealy' macT 0 inp
 @
 
-Where the first argument of @'mealy'@ is our @macT@ function, and the second
+Where the first argument of 'mealy' is our @macT@ function, and the second
 argument is the initial state, in this case 0. We can see it is functioning
 correctly in our interpreter:
 
@@ -505,7 +515,7 @@ macT acc (x,y) = (acc',o)
     acc' = ma acc (x,y)
     o    = acc
 
-mac inp = 'mealy' macT 0 inp
+mac xy = 'mealy' macT 0 xy
 
 topEntity
   :: 'Clock' 'System'
@@ -642,8 +652,9 @@ structure.
     function directly:
 
     @
-    macN (x,y) = acc
+    macN xy = acc
       where
+        (x,y) = 'unbundle' xy
         acc = 'register' 0 (acc + x * y)
     @
 
@@ -655,52 +666,47 @@ structure.
     class:
 
     @
-    macA (x,y) = acc
+    macA xy = acc
       where
         acc  = 'register' 0 acc'
-        acc' = ma '<$>' acc '<*>' 'bundle' (x,y)
+        acc' = ma '<$>' acc '<*>' xy
     @
 
-* __'Control.Monad.State.Lazy.State' Monad__
+* __t'Control.Monad.State.Strict.State' Monad__
 
-    We can also implement the original @macT@ function as a
-    @'Control.Monad.State.Lazy.State'@
-    monadic computation. First we must add an extra import statement, right
-    after the import of "Clash.Prelude":
+    We can also implement the original @macT@ function as
+    a t'Control.Monad.State.Strict.State' monadic computation. First we must add
+    an extra import statement, right after the import of "Clash.Prelude":
 
     @
-    import Control.Monad.State
+    import Control.Monad.State.Strict
     @
 
     We can then implement macT as follows:
 
     @
     macTS (x,y) = do
-      acc <- 'Control.Monad.State.Lazy.get'
-      'Control.Monad.State.Lazy.put' (acc + x * y)
+      acc <- 'Control.Monad.State.Strict.get'
+      'Control.Monad.State.Strict.put' (acc + x * y)
       return acc
     @
 
-    We can use the 'mealy' function again, although we will have to change
-    position of the arguments and result:
+    We can use the 'mealyS' function to run our stateful implementation, this
+    can simplify translating algorithms which are described imperatively.
 
     @
-    asStateM
-      :: ( 'HiddenClockResetEnable' dom
-         , 'NFDataX' s )
-      => (i -> 'Control.Monad.State.Lazy.State' s o)
+    mealyS
+      :: ( HiddenClockResetEnable dom, NFDataX s )
+      => (i -> State s o)
       -> s
-      -> ('Signal' dom i -> 'Signal' dom o)
-    asStateM f i = 'mealy' g i
-      where
-        g s x = let (o,s') = 'Control.Monad.State.Lazy.runState' (f x) s
-                in  (s',o)
+      -> (Signal dom i -> Signal dom o)
+    mealyS f initS = ...
     @
 
     We can then create the complete @mac@ circuit as:
 
     @
-    macS = asStateM macTS 0
+    macS xy = 'mealyS' macTS 0 xy
     @
 -}
 
@@ -906,19 +912,18 @@ import "Clash.Signal"
 import "Clash.Prelude"
 import "Clash.Intel.ClockGen"
 
-'createDomain' vSystem{vName=\"DomInput\", vPeriod=20000}
+'createDomain' vSystem{vName=\"DomInput\", vPeriod=20000, vResetPolarity=ActiveLow}
 'createDomain' vSystem{vName=\"Dom100\", vPeriod=10000}
 
 topEntity
   :: Clock DomInput
-  -> Signal DomInput Bool
+  -> Reset DomInput
   -> Signal Dom100 Bit
   -> Signal Dom100 (BitVector 8)
 topEntity clk rst =
-    'exposeClockResetEnable' ('mealy' blinkerT (1,False,0) . Clash.Prelude.isRising 1) pllOut rstSync 'enableGen'
+    'exposeClockResetEnable' ('mealy' blinkerT (1,False,0) . 'Clash.Prelude.isRising' 1) pllOut pllRst 'enableGen'
   where
-    (pllOut,pllStable) = 'Clash.Intel.ClockGen.altpll' \@Dom100 (SSymbol \@\"altpll100\") clk ('Clash.Signal.unsafeFromLowPolarity' rst)
-    rstSync            = 'Clash.Signal.resetSynchronizer' pllOut ('Clash.Signal.unsafeFromLowPolarity' pllStable)
+    (pllOut,pllRst) = 'Clash.Intel.ClockGen.altpllSync' clk rst
 
 blinkerT (leds,mode,cntr) key1R = ((leds',mode',cntr'),leds)
   where
@@ -952,8 +957,9 @@ use work.Blinker_topEntity_types.all;
 entity topEntity is
   port(-- clock
        clk    : in Blinker_topEntity_types.clk_DomInput;
-       rst    : in boolean;
-       x      : in std_logic;
+       -- reset
+       rst    : in Blinker_topEntity_types.rst_DomInput;
+       eta    : in std_logic;
        result : out std_logic_vector(7 downto 0));
 end;
 
@@ -987,8 +993,9 @@ use work.blinker_types.all;
 
 entity blinker is
   port(-- clock
-       CLOCK_50 : in blinker_types.clk_dominput;
-       KEY0     : in boolean;
+       CLOCK_50 : in blinker_types.clk_DomInput;
+       -- reset
+       KEY0     : in blinker_types.rst_DomInput;
        KEY1     : in std_logic;
        LED      : out std_logic_vector(7 downto 0));
 end;
@@ -1043,7 +1050,7 @@ primitives, using 'Signed' multiplication (@*@) as an example. The
 "Clash.Sized.Internal.Signed" module specifies multiplication as follows:
 
 @
-(*#) :: 'GHC.TypeLits.KnownNat' n => 'Signed' n -> 'Signed' n -> 'Signed' n
+(*#) :: 'KnownNat' n => 'Signed' n -> 'Signed' n -> 'Signed' n
 (S a) *# (S b) = fromInteger_INLINE (a * b)
 {\-\# NOINLINE (*#) \#-\}
 @
@@ -1102,7 +1109,7 @@ blockRam#
   :: ( KnownDomain dom
      , HasCallStack
      , NFDataX a )
-  => 'Clock' dom           -- ^ 'Clock' to synchronize to
+  => 'Clock' dom           -- ^ Clock to synchronize to
   -> 'Enable' dom          -- ^ Global enable
   -> 'Vec' n a             -- ^ Initial content of the BRAM, also
                            -- determines the size, @n@, of the BRAM.
@@ -1255,7 +1262,7 @@ a general listing of the available template holes:
   @-fclash-vivado@ flag. To be used with in an @~IF .. ~THEN .. ~ELSE .. ~FI@
   statement.
 * @~CMPLE[\<HOLE1\>][\<HOLE2\>]@: /1/ when @\<HOLE1\> \<= \<HOLE2\>@, otherwise /0/
-* @~IW64@: /1/ when Int/Word/Integer types are represented with 64 bits in HDL.
+* @~IW64@: /1/ when Int\/Word\/Integer types are represented with 64 bits in HDL.
   /0/ when they're represented by 32 bits.
 * @~TOBV[\<HOLE\>][\<TYPE\>]@: create conversion code that so that the
   expression in @\<HOLE\>@ is converted to a bit vector (@std_logic_vector@).
@@ -1282,16 +1289,16 @@ a general listing of the available template holes:
 * @~TAG[N]@: Name of given domain. Errors when called on an argument which is not
   a 'KnownDomain', 'Reset', or 'Clock'.
 * @~PERIOD[N]@: Clock period of given domain. Errors when called on an argument
-  which is not a 'KnownDomain' or 'KnownConf'.
+  which is not a 'Clock', 'Reset', 'KnownDomain' or 'KnownConf'.
 * @~ISACTIVEENABLE[N]@: Is the @(N+1)@'th argument a an Enable line NOT set to a
   constant True. Can be used instead of deprecated (and removed) template tag
 * @~ISSYNC[N]@: Does synthesis domain at the @(N+1)@'th argument have synchronous resets. Errors
-  when called on an argument which is not a 'KnownDomain' or 'KnownConf'.
+  when called on an argument which is not a 'Reset', 'Clock', 'Enable', 'KnownDomain' or 'KnownConf'.
 * @~ISINITDEFINED[N]@: Does synthesis domain at the @(N+1)@'th argument have defined initial
-  values. Errors when called on an argument which is not a 'KnownDomain' or 'KnownConf'.
+  values. Errors when called on an argument which is not a 'Clock', 'Reset', 'Enable', 'KnownDomain' or 'KnownConf'.
 * @~ACTIVEEDGE[edge][N]@: Does synthesis domain at the @(N+1)@'th argument respond to
   /edge/. /edge/ must be one of 'Falling' or 'Rising'. Errors when called on an
-  argument which is not a 'KnownDomain' or 'KnownConf'.
+  argument which is not a 'Clock', 'Reset', 'Enable', 'KnownDomain' or 'KnownConf'.
 * @~AND[\<HOLE1\>,\<HOLE2\>,..]@: Logically /and/ the conditions in the @\<HOLE\>@'s
 * @~VAR[\<NAME\>][N]@: Like @~ARG[N]@ but binds the argument to a variable named NAME.
   The @\<NAME\>@ can be left blank, then clash will come up with a (unique) name.
@@ -1556,9 +1563,10 @@ __asyncRam__
      , 'HasCallStack'
      , 'KnownDomain' wdom
      , 'KnownDomain' rdom
+     , 'NFDataX' a
      )
-  => 'Clock' wdom                     -- ^ 'Clock' to which to synchronize the write port of the RAM
-  -> 'Clock' rdom                     -- ^ 'Clock' to which the read address signal, @r@, is synchronized to
+  => 'Clock' wdom                     -- ^ Clock to which to synchronize the write port of the RAM
+  -> 'Clock' rdom                     -- ^ Clock to which the read address signal, @r@, is synchronized to
   -> 'Enable' wdom                    -- ^ Global enable
   -> 'SNat' n                         -- ^ Size @n@ of the RAM
   -> 'Signal' rdom addr               -- ^ Read address @r@
@@ -1657,12 +1665,10 @@ ptrSync clk1 clk2 rst2 en2 =
   'Clash.Explicit.Signal.register' clk2 rst2 en2 0 . 'Clash.Explicit.Signal.register' clk2 rst2 en2 0 . 'Clash.Explicit.Signal.unsafeSynchronizer' clk1 clk2
 @
 
-It uses the 'Clash.Explicit.Signal.unsafeSynchronizer' primitive, which is needed to go from one clock
-domain to the other. All synchronizers are specified in terms of
-'Clash.Explicit.Signal.unsafeSynchronizer' (see for example the <src/Clash-Prelude-RAM.html#line-103 source of asyncRam>).
-The 'Clash.Explicit.Signal.unsafeSynchronizer' primitive is turned into a (bundle of) wire(s) by the
-Clash compiler, so developers must ensure that it is only used as part of a
-proper synchronizer.
+It uses the 'Clash.Explicit.Signal.unsafeSynchronizer' primitive, which is
+needed to go from one clock domain to the other. The 'Clash.Explicit.Signal.unsafeSynchronizer'
+primitive is turned into a (bundle of) wire(s) by the Clash compiler, so
+developers must ensure that it is only used as part of a proper synchronizer.
 
 Finally we combine all the components in:
 
@@ -1675,9 +1681,9 @@ asyncFIFOSynchronizer
   -- ^ Size of the internally used addresses, the  FIFO contains @2^addrSize@
   -- elements.
   -> 'Clock' wdom
-  -- ^ 'Clock' to which the write port is synchronized
+  -- ^ Clock to which the write port is synchronized
   -> 'Clock' rdom
-  -- ^ 'Clock' to which the read port is synchronized
+  -- ^ Clock to which the read port is synchronized
   -> 'Reset' wdom
   -> 'Reset' rdom
   -> 'Enable' wdom
@@ -2193,7 +2199,7 @@ Here is a list of Haskell features for which the Clash compiler has only
             __plusFloat#__ :: 'Float#' -> 'Float#' -> 'Float#'
             @
 
-            which underlie @'Float'@'s 'Num' instance, must be implemented as
+            which underlie 'Float'\'s 'Num' instance, must be implemented as
             purely combinational circuits according to their type. Remember,
             sequential circuits operate on values of type \"@'Signal' a@\".
 
@@ -2294,9 +2300,9 @@ function. This enables the following features not available to Lava:
 * Automatic synthesis for user-defined ADTs
 * Synthesis of all choice constructs (pattern matching, guards, etc.)
 * 'Applicative' instance for the 'Signal' type
-* Working with \"normal\" functions permits the use of e.g. the
-  'Control.Monad.State.Lazy.State' monad to describe the functionality of a
-  circuit.
+* Working with \"normal\" functions permits the use of e.g.
+  the t'Control.Monad.State.Strict.State' monad to describe the functionality of
+  a circuit.
 
 Although there are Lava alternatives to some of the above features (e.g.
 first-class patterns to replace pattern matching) they are not as \"beautiful\"
@@ -2374,7 +2380,7 @@ testBench :: Signal System Bool
 testBench = done
   where
     testInput      = stimuliGenerator clk rst (2:>3:>(-2):>8:>Nil)
-    expectedOutput = outputVerifier clk rst (4:>12:>1:>20:>Nil)
+    expectedOutput = outputVerifier' clk rst (4:>12:>1:>20:>Nil)
     done           = expectedOutput (topEntity clk rst testInput)
     clk            = tbSystemClockGen (not \<$\> done)
     rst            = systemResetGen
@@ -2498,8 +2504,8 @@ topEntity
 topEntity clk rst =
     'exposeClockResetEnable' ('mealy' blinkerT (1,False,0) . Clash.Prelude.isRising 1) pllOut rstSync 'enableGen'
   where
-    (pllOut,pllStable) = 'Clash.Intel.ClockGen.altpll' \@Dom100 (SSymbol \@\"altpll100\") clk ('Clash.Signal.unsafeFromLowPolarity' rst)
-    rstSync            = 'Clash.Signal.resetSynchronizer' pllOut ('Clash.Signal.unsafeFromLowPolarity' pllStable)
+    (pllOut,pllStable) = 'Clash.Intel.ClockGen.altpll' \@Dom100 (SSymbol \@\"altpll100\") clk ('unsafeFromActiveLow' rst)
+    rstSync            = 'Clash.Signal.resetSynchronizer' pllOut ('unsafeFromActiveLow' pllStable)
 
 blinkerT (leds,mode,cntr) key1R = ((leds',mode',cntr'),leds)
   where
