@@ -2,7 +2,7 @@
 Copyright  :  (C) 2016,      University of Twente,
                   2017,      QBayLogic, Google Inc.
                   2017-2019, Myrtle Software Ltd,
-                  2021-2022, QBayLogic B.V.
+                  2021-2023, QBayLogic B.V.
 License    :  BSD2 (see the file LICENSE)
 Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -59,6 +59,9 @@ import           Data.Functor.Identity (Identity)
 import           Data.Functor.Product (Product)
 import           Data.Functor.Sum    (Sum)
 import           Data.Int            (Int8, Int16, Int32, Int64)
+import qualified Data.List.Infinite  as Inf
+import           Data.List.Infinite  (Infinite (..))
+import           Data.List.NonEmpty  (NonEmpty)
 import           Data.Ord            (Down (Down))
 import           Data.Ratio          (Ratio, numerator, denominator)
 import qualified Data.Semigroup      as SG
@@ -97,6 +100,8 @@ infixr 0 `defaultSeqX`
 -- out with an exception.
 errorX :: HasCallStack => String -> a
 errorX msg = throw (XException ("X: " ++ msg ++ "\n" ++ prettyCallStack callStack))
+{-# NOINLINE errorX #-}
+{-# ANN errorX hasBlackBox #-}
 
 -- | Convert 'XException' to 'ErrorCall'
 --
@@ -122,7 +127,7 @@ errorX msg = throw (XException ("X: " ++ msg ++ "\n" ++ prettyCallStack callStac
 -- > {-# LANGUAGE ViewPatterns, BangPatterns #-}
 -- > f (xToErrorCtx "a is X" -> !a) (xToErrorCtx "b is X" -> !b) = ...
 --
--- __NB:__ Fully synthesizable, so doesn't have to be removed before synthesis
+-- __NB__: Fully synthesizable, so doesn't have to be removed before synthesis
 --
 -- === __Example__
 --
@@ -148,7 +153,8 @@ xToErrorCtx ctx a = unsafeDupablePerformIO
   (catch (evaluate a >> return a)
          (\(XException msg) ->
            throw (ErrorCall (unlines [ctx,msg]))))
-{-# NOINLINE xToErrorCtx #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE xToErrorCtx #-}
 
 -- | Convert 'XException' to 'ErrorCall'
 --
@@ -176,7 +182,7 @@ xToErrorCtx ctx a = unsafeDupablePerformIO
 -- > {-# LANGUAGE ViewPatterns, BangPatterns #-}
 -- > f (xToError -> !a) (xToError -> !b) = ...
 --
--- __NB:__ Fully synthesizable, so doesn't have to be removed before synthesis
+-- __NB__: Fully synthesizable, so doesn't have to be removed before synthesis
 --
 -- === __Example__
 --
@@ -218,7 +224,8 @@ xToError = xToErrorCtx (prettyCallStack callStack)
 seqX :: a -> b -> b
 seqX a b = unsafeDupablePerformIO
   (catch (evaluate a >> return b) (\(XException _) -> return b))
-{-# NOINLINE seqX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE seqX #-}
 {-# ANN seqX hasBlackBox #-}
 infixr 0 `seqX`
 
@@ -234,7 +241,8 @@ seqErrorX a b = unsafeDupablePerformIO
      [ Handler (\(XException _) -> return b)
      , Handler (\(ErrorCall _) -> return b)
      ])
-{-# NOINLINE seqErrorX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE seqErrorX #-}
 {-# ANN seqErrorX hasBlackBox #-}
 infixr 0 `seqErrorX`
 
@@ -245,7 +253,7 @@ infixr 0 `seqErrorX`
 -- output of such a component (typically a unit) can be passed as the first
 -- argument to 'hwSeqX' to ensure the ILA ends up in the generated HDL.
 --
--- __NB__: the result of 'hwSeqX' must (indirectly) be used at the very top of
+-- __NB__: The result of 'hwSeqX' must (indirectly) be used at the very top of
 -- a design. If it's not, Clash will remove it like it does for any other unused
 -- circuit parts.
 --
@@ -253,18 +261,22 @@ infixr 0 `seqErrorX`
 -- uses 'Clash.Netlist.BlackBox.Types.RenderVoid'
 hwSeqX :: a -> b -> b
 hwSeqX = seqX
-{-# NOINLINE hwSeqX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE hwSeqX #-}
 {-# ANN hwSeqX hasBlackBox #-}
 infixr 0 `hwSeqX`
 
 -- | Evaluate a value with given function, returning 'Nothing' if it throws
--- 'XException'.
+-- 'XException'. Note that non-'XException' errors take precedence over 'XException'
+-- ones
 --
--- > maybeX hasX 42                  = Just 42
--- > maybeX hasX (XException msg)    = Nothing
--- > maybeX hasX (3, XException msg) = Nothing
--- > maybeX hasX (3, _|_)            = _|_
--- > maybeX hasX _|_                 = _|_
+-- > maybeX hasX 42                    = Just 42
+-- > maybeX hasX (XException msg)      = Nothing
+-- > maybeX hasX (3, XException msg)   = Nothing
+-- > maybeX hasX (XException msg, _|_) = _|_
+-- > maybeX hasX (_|_, XException msg) = _|_
+-- > maybeX hasX (3, _|_)              = _|_
+-- > maybeX hasX _|_                   = _|_
 -- >
 -- > maybeX isX 42                  = Just 42
 -- > maybeX isX (XException msg)    = Nothing
@@ -275,15 +287,18 @@ infixr 0 `hwSeqX`
 maybeX :: (a -> Either String a) -> a -> Maybe a
 maybeX f a = either (const Nothing) Just (f a)
 
--- | Fully evaluate a value, returning 'Nothing' if it throws 'XException'.
+-- | Fully evaluate a value, returning 'Nothing' if it throws 'XException'. Note
+-- that non-'XException' errors take precedence over 'XException' ones.
 --
--- > maybeHasX 42                  = Just 42
--- > maybeHasX (XException msg)    = Nothing
--- > maybeHasX (3, XException msg) = Nothing
--- > maybeHasX (3, _|_)            = _|_
--- > maybeHasX _|_                 = _|_
+-- > maybeHasX 42                    = Just 42
+-- > maybeHasX (XException msg)      = Nothing
+-- > maybeHasX (3, XException msg)   = Nothing
+-- > maybeHasX (XException msg, _|_) = _|_
+-- > maybeHasX (_|_, XException msg) = _|_
+-- > maybeHasX (3, _|_)              = _|_
+-- > maybeHasX _|_                   = _|_
 --
-maybeHasX :: NFData a => a -> Maybe a
+maybeHasX :: (NFData a, NFDataX a) => a -> Maybe a
 maybeHasX = maybeX hasX
 
 -- | Evaluate a value to WHNF, returning 'Nothing' if it throws 'XException'.
@@ -300,21 +315,28 @@ maybeIsX = maybeX isX
 -- If you want to determine if a value contains undefined parts, use
 -- 'hasUndefined' instead.
 --
--- > hasX 42                  = Right 42
--- > hasX (XException msg)    = Left msg
--- > hasX (3, XException msg) = Left msg
--- > hasX (3, _|_)            = _|_
--- > hasX _|_                 = _|_
+-- > hasX 42                    = Right 42
+-- > hasX (XException msg)      = Left msg
+-- > hasX (3, XException msg)   = Left msg
+-- > hasX (XException msg, _|_) = _|_
+-- > hasX (_|_, XException msg) = _|_
+-- > hasX (3, _|_)              = _|_
+-- > hasX _|_                   = _|_
 --
 -- If a data structure contains multiple 'XException's, the "first" message is
--- picked according to the implementation of 'rnf'.
-hasX :: NFData a => a -> Either String a
+-- picked according to the implementation of 'rnfX'.
+hasX :: (NFData a, NFDataX a) => a -> Either String a
 hasX a =
+  -- TODO: Whenever 'a' contains an 'XException', we need to reevaluate the
+  --       structure using 'rnfX' to make sure it didn't also contain another
+  --       error call. We could prevent the two traversals by making 'hasX' a
+  --       type class method. Also see: https://github.com/clash-lang/clash-compiler/issues/2450.
   unsafeDupablePerformIO
     (catch
       (evaluate (rnf a) >> return (Right a))
-      (\(XException msg) -> return (Left msg)))
-{-# NOINLINE hasX #-}
+      (\(XException msg) -> evaluate (rnfX a) >> return (Left msg)))
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE hasX #-}
 
 -- | Evaluate a value to WHNF, returning @'Left' msg@ if is a 'XException'.
 --
@@ -329,7 +351,8 @@ isX a =
     (catch
       (evaluate a >> return (Right a))
       (\(XException msg) -> return (Left msg)))
-{-# NOINLINE isX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE isX #-}
 
 -- | Like the 'Show' class, but values that normally throw an 'XException' are
 -- converted to @undefined@, instead of error'ing out with an exception.
@@ -420,6 +443,8 @@ instance ShowX Integer where
 instance ShowX Natural where
   showsPrecX = showsPrecXWith showsPrec
 
+instance ShowX Ordering
+
 instance ShowX a => ShowX (Seq a) where
   showsPrecX _ = showListX . toList
 
@@ -460,7 +485,8 @@ forceX x = x `deepseqX` x
 -- second. Does not propagate 'XException's.
 deepseqX :: NFDataX a => a -> b -> b
 deepseqX a b = rnfX a `seq` b
-{-# NOINLINE deepseqX #-}
+-- See: https://github.com/clash-lang/clash-compiler/pull/2511
+{-# CLASH_OPAQUE deepseqX #-}
 {-# ANN deepseqX hasBlackBox #-}
 infixr 0 `deepseqX`
 
@@ -538,8 +564,25 @@ instance NFDataX a => NFDataX (Down a) where
   hasUndefined d@(~(Down x))= if isLeft (isX d) then True else hasUndefined x
   ensureSpine ~(Down x) = Down (ensureSpine x)
 
+instance NFDataX a => NFDataX (Infinite a) where
+  deepErrorX msg = Inf.repeat (deepErrorX msg)
+  rnfX d@(~(x :< xs)) =
+    if isLeft (isX d) then
+      ()
+    else
+      rnfX x `seq` rnfX xs
+  hasUndefined d@(~(x :< xs)) =
+    if isLeft (isX d) then
+      True
+    else
+      hasUndefined x || hasUndefined xs
+
+  ensureSpine ~(x :< xs) = ensureSpine x :< ensureSpine xs
+
 instance NFDataX Bool
+instance NFDataX Ordering
 instance NFDataX a => NFDataX [a]
+instance NFDataX a => NFDataX (NonEmpty a)
 instance (NFDataX a, NFDataX b) => NFDataX (Either a b)
 instance NFDataX a => NFDataX (Maybe a)
 instance NFDataX a => NFDataX (Identity a)
@@ -695,13 +738,13 @@ instance NFDataX a => NFDataX (M.Last a)
 instance NFDataX a => NFDataX (SG.Option a)
 #endif
 
--- | __N.B.__: The documentation only shows instances up to /3/-tuples. By
+-- | __NB__: The documentation only shows instances up to /3/-tuples. By
 -- default, instances up to and including /12/-tuples will exist. If the flag
 -- @large-tuples@ is set instances up to the GHC imposed limit will exist. The
 -- GHC imposed limit is either 62 or 64 depending on the GHC version.
 mkShowXTupleInstances [2..maxTupleSize]
 
--- | __N.B.__: The documentation only shows instances up to /3/-tuples. By
+-- | __NB__: The documentation only shows instances up to /3/-tuples. By
 -- default, instances up to and including /12/-tuples will exist. If the flag
 -- @large-tuples@ is set instances up to the GHC imposed limit will exist. The
 -- GHC imposed limit is either 62 or 64 depending on the GHC version.
@@ -713,6 +756,6 @@ undefined = errorX "undefined"
 
 -- | Same as 'Data.Maybe.fromJust', but returns a bottom/undefined value that
 -- other Clash constructs are aware of.
-fromJustX :: HasCallStack => Maybe a -> a
-fromJustX Nothing = errorX "fromJustX: Nothing"
+fromJustX :: (HasCallStack, NFDataX a) => Maybe a -> a
+fromJustX Nothing = deepErrorX "fromJustX: Nothing"
 fromJustX (Just a) = a
